from save_to_db.core.item import Item
from save_to_db.utils.test_base import TestBase
from save_to_db.exceptions import ItemsNotTheSame, MultipleItemsMatch


class TestSimpleConfig(TestBase):
    
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    
    #--- batch_size ------------------------------------------------------------
    
    def test_batch_size(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
        
        # need this for `ItemGeneralOne` (related field)
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
        
        persister = self.persister
        db_adapter = persister.db_adapter
        original_adapter_batch_size = db_adapter.BATCH_SIZE
        original_adapter_get = db_adapter.get
        list_of_get_sizes = []
        def decorated_adapter_get(items_and_fkeys, *args, **kwargs):
            nonlocal list_of_get_sizes
            list_of_get_sizes.append(len(items_and_fkeys))
            return original_adapter_get(items_and_fkeys, *args, **kwargs)
        db_adapter.get = decorated_adapter_get
        
        try:
            # unlimited batch size
            db_adapter.BATCH_SIZE = 1000
            bulk = ItemGeneralOne.Bulk()
            for i in range(10):
                bulk.gen(f_integer=i)
            persister.persist(bulk)
            # got everything in one batch
            self.assertEqual(list_of_get_sizes, [10])
            list_of_get_sizes.clear()
            
            db_adapter.BATCH_SIZE = 3
            persister.persist(bulk)
            # 4 batches with max size of 3
            self.assertEqual(list_of_get_sizes, [3, 3, 3, 1])
            list_of_get_sizes.clear()
            
            # unlimited again
            db_adapter.BATCH_SIZE = 1000
            persister.persist(bulk)
            self.assertEqual(list_of_get_sizes, [10])
            list_of_get_sizes.clear()
            
            # limit on item class side
            ItemGeneralOne.batch_size = 4
            persister.persist(bulk)
            # 2 batches with max size of 4
            self.assertEqual(list_of_get_sizes, [4, 4, 2])
            list_of_get_sizes.clear()
            
            bulk = ItemGeneralTwo.Bulk()
            for i in range(10):
                bulk.gen(f_integer=i)
            persister.persist(bulk)
            # second class not effected
            self.assertEqual(list_of_get_sizes, [10])
            list_of_get_sizes.clear()
        finally:
            db_adapter.get = original_adapter_get
            db_adapter.BATCH_SIZE = original_adapter_batch_size
    
    
    #--- nullables -------------------------------------------------------------
    
    def test_nullables(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
            nullables = ['f_text', 'two_1_1', 'two_x_1', 'two_x_x']
            relations = {
                'two_x_x': {
                    'replace_x_to_many': True,
                }
            }
         
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
        
        persister = self.persister
        
        item_one = ItemGeneralOne(
            f_integer=1,
            f_text='text',
            two_1_1=ItemGeneralTwo(f_integer=2),
            two_x_1=ItemGeneralTwo(f_integer=3),
            two_x_x=[ItemGeneralTwo(f_integer=4)],
        )
        
        persister.persist(item_one)
        model_ones = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(model_ones), 1)
        model_one = model_ones[0]
        self.assertEqual(model_one.f_integer, 1)
        self.assertEqual(model_one.f_text, 'text')
        self.assertEqual(model_one.two_1_1.f_integer, 2)
        self.assertEqual(model_one.two_x_1.f_integer, 3)
        two_x_x = self.get_related_x_to_many(model_one, 'two_x_x')
        self.assertEqual(len(two_x_x), 1)
        self.assertEqual(two_x_x[0].f_integer, 4)
        
        # rewriting nullables with other values
        item_one = ItemGeneralOne(
            f_integer=1,
            f_text='text-2',
            two_1_1=ItemGeneralTwo(f_integer=1001),
            two_x_1=ItemGeneralTwo(f_integer=1002),
            two_x_x=[ItemGeneralTwo(f_integer=1003),
                     ItemGeneralTwo(f_integer=1004)],
        )
        persister.persist(item_one)
        model_ones = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(model_ones), 1)
        model_one = model_ones[0]
        self.assertEqual(model_one.f_integer, 1)
        self.assertEqual(model_one.f_text, 'text-2')
        self.assertEqual(model_one.two_1_1.f_integer, 1001)
        self.assertEqual(model_one.two_x_1.f_integer, 1002)
        two_x_x = self.get_related_x_to_many(model_one, 'two_x_x')
        self.assertEqual(len(two_x_x), 2)
        self.assertEqual(two_x_x[0].f_integer, 1003)
        self.assertEqual(two_x_x[1].f_integer, 1004)
        
        # no values fo nullables
        item_one = ItemGeneralOne(f_integer=1)
        persister.persist(item_one)
        model_ones = self.get_all_models(self.ModelGeneralOne)
        self.assertEqual(len(model_ones), 1)
        model_one = model_ones[0]
        self.assertEqual(model_one.f_integer, 1)
        self.assertIsNone(model_one.f_text)
        self.assertTrue(not hasattr(model_one, 'two_1_1') or
                        model_one.two_1_1 is None)
        two_x_x = self.get_related_x_to_many(model_one, 'two_x_x')
        self.assertEqual(len(two_x_x), 0)
    
    
    #--- fast_insert -----------------------------------------------------------
    
    def test_fast_insert(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = ['f_integer']
            getters = ['f_integer']
            fast_insert = True
            
        persister = self.persister
        sort_func = lambda model: (model.f_integer, model.f_string)
        
        #--- no fast_insert ---
        item_one = ItemGeneralOne(f_integer=1, f_string='one')
        persister.persist(item_one)
        
        bulk_one = ItemGeneralOne.Bulk()
        bulk_one.gen(f_integer=1, f_string='one-updated')
        bulk_one.gen(f_integer=2, f_string='two')
        persister.persist(bulk_one)
        
        model_ones = self.get_all_models(self.ModelGeneralOne,
                                         sort_key=sort_func)
        self.assertEqual(len(model_ones), 2)
        self.assertEqual(model_ones[0].f_string, 'one-updated')
        self.assertEqual(model_ones[1].f_string, 'two')
        
        #--- fast_insert ---
        item_two = ItemGeneralTwo(f_integer=1, f_string='one')
        persister.persist(item_two)
        
        bulk_two = ItemGeneralTwo.Bulk()
        bulk_two.gen(f_integer=1, f_string='one-updated')
        bulk_two.gen(f_integer=2, f_string='two')
        # although "f_integer" field is a getter, since we use fast_insert,
        # new model will be create without checking for update
        persister.persist(bulk_two)
        
        model_twos = self.get_all_models(self.ModelGeneralTwo,
                                         sort_key=sort_func)
        self.assertEqual(len(model_twos), 3)
        self.assertEqual(model_twos[0].f_integer, 1)
        self.assertEqual(model_twos[0].f_string, 'one')
        self.assertEqual(model_twos[1].f_integer, 1)
        self.assertEqual(model_twos[1].f_string, 'one-updated')
        self.assertEqual(model_twos[2].f_integer, 2)
        self.assertEqual(model_twos[2].f_string, 'two')
    
    
    #--- get_only_mode ---------------------------------------------------------
    
    def test_get_only_mode(self):
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}, {'two_x_x'}]
            
            def __str__(self):
                return 'ItemGeneralOne, f_integer: {}'.format(self['f_integer'])
            
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
            
            def __str__(self):
                return 'ItemGeneralTwo, f_integer: {}'.format(self['f_integer'])
        
        persister = self.persister
        get_key = lambda model: (model.f_integer,
                                 model.f_string if model.f_string else '')
        
        #--- normal model creating ---------------------------------------------
        bulk_item_one = ItemGeneralOne.Bulk()
        
        for i in range(2):
            bulk_item_one.gen(f_integer=i,
                              two_x_x=[ItemGeneralTwo(f_integer=i)])
        
        self.assertFalse(ItemGeneralOne.get_only_mode)
        self.assertFalse(ItemGeneralTwo.get_only_mode)
        
        persister.persist(bulk_item_one)
        model_ones = self.get_all_models(self.ModelGeneralOne, sort_key=get_key)
        
        self.assertEqual(len(model_ones), 2)
        self.assertEqual(model_ones[0].f_integer, 0)
        self.assertEqual(model_ones[1].f_integer, 1)
        
        two_x_x_first = self.get_related_x_to_many(model_ones[0], 'two_x_x',
                                                   sort_key=get_key)
        self.assertEqual(len(two_x_x_first), 1)
        self.assertEqual(two_x_x_first[0].f_integer, 0)
        
        two_x_x_second = self.get_related_x_to_many(model_ones[1], 'two_x_x',
                                                    sort_key=get_key)
        self.assertEqual(len(two_x_x_second), 1)
        self.assertEqual(two_x_x_second[0].f_integer, 1)
        
        #--- some items in get_only_mode ---------------------------------------
        bulk_item_one = ItemGeneralOne.Bulk()
        
        # get only (cannot be created)
        # (related by "two_x_x" field item must be created)
        not_saved_item = bulk_item_one.gen(
            f_integer=10, f_string='str-{}'.format(10),
            two_x_x=[ItemGeneralTwo(f_integer=10)])
        not_saved_item.get_only_mode = True
        
        # normal
        item = bulk_item_one.gen(f_integer=20,
                                 f_string='str-{}'.format(20),
                                 two_x_x=[ItemGeneralTwo(f_integer=20)])
        
        
        # old items
        for i in range(2):
            item = bulk_item_one.gen(
                f_integer=i, f_string='str-{}'.format(i),
                two_x_x=[ItemGeneralTwo(f_integer=i,
                                        f_string='str-{}'.format(i))])
            
            # second item update only
            if i == 1:
                del item['f_integer']  # forcing to use 'two_x_x' getter
                item.get_only_mode = True
                item['two_x_x'][0].get_only_mode = True
        
        
        saved_items, _ = persister.persist(bulk_item_one)
        
        # f_integer values are 0, 1, 20 (with f_integer=1 deleted)    
        self.assertEqual(len(saved_items), 3)
        self.assertNotIn(not_saved_item, saved_items)
        
        # all items can be loaded except for `f_integer=10` one
        for item in bulk_item_one:
            if item is not_saved_item:
                continue
            self.assertIn(item, saved_items)
        
        # checking the data
        model_ones = self.get_all_models(self.ModelGeneralOne, sort_key=get_key)
        self.assertEqual(len(model_ones), 3)
        
        self.assertEqual(model_ones[0].f_integer, 0)
        self.assertEqual(model_ones[1].f_integer, 1)
        self.assertEqual(model_ones[2].f_integer, 20)
        
        self.assertEqual(model_ones[0].f_string, 'str-0')
        self.assertIsNone(model_ones[1].f_string)  # get_only_mode
        self.assertEqual(model_ones[2].f_string, 'str-20')
        
        # checking model_two without model_one
        model_twos = self.get_all_models(self.ModelGeneralTwo, sort_key=get_key)
        
        self.assertEqual(len(model_twos), 4)
        model_two = model_twos[-2]  # with `f_integer=10`
        self.assertEqual(model_two.f_integer, 10)
        
        one_x_x = self.get_related_x_to_many(model_two, 'one_x_x')
        self.assertEqual(len(one_x_x), 0)  # get_only_mode
        
        # checking all other model_two
        two_x_x = self.get_related_x_to_many(model_ones[0], 'two_x_x',
                                             sort_key=get_key)
        self.assertEqual(len(two_x_x), 1)
        self.assertEqual(two_x_x[0].f_integer, 0)
        self.assertEqual(two_x_x[0].f_string, 'str-0')
        
        two_x_x = self.get_related_x_to_many(model_ones[1], 'two_x_x',
                                             sort_key=get_key)
        self.assertEqual(len(two_x_x), 1)
        self.assertEqual(two_x_x[0].f_integer, 1)
        self.assertIsNone(two_x_x[0].f_string)  # get_only_mode
        
        two_x_x = self.get_related_x_to_many(model_ones[2], 'two_x_x',
                                             sort_key=get_key)
        self.assertEqual(len(two_x_x), 1)
        self.assertEqual(two_x_x[0].f_integer, 20)
    
    
    #--- update_only_mode ------------------------------------------------------
    
    def test_update_only_mode(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}]
            update_only_mode = True
        
        self.item_cls_manager.autogenerate = True
        persister = self.persister
        
        bulk_one = ItemGeneralOne.Bulk()
        item = bulk_one.gen(f_integer='10', f_text='10-first')
        item.update_only_mode = False
        item = bulk_one.gen(f_integer='20', f_text='20-first')
        item.update_only_mode = False
        
        items, _ = persister.persist(bulk_one)
        self.assertEqual(len(items), 2)
        
        sort_func = lambda model: model.f_integer
        
        models_one = self.get_all_models(self.ModelGeneralOne,
                                         sort_key=sort_func)
        self.assertEqual(len(models_one), 2)
        self.assertEqual(models_one[0].f_integer, 10)
        self.assertEqual(models_one[0].f_text, '10-first')
        self.assertEqual(models_one[1].f_integer, 20)
        self.assertEqual(models_one[1].f_text, '20-first')
        
        bulk_one = ItemGeneralOne.Bulk()
        item = bulk_one.gen(f_integer='10', f_text='10-second')
        item = bulk_one.gen(f_integer='30', f_text='30-second')
        
        items, _ = persister.persist(bulk_one)
        self.assertEqual(len(items), 1)
        models_one = self.get_all_models(self.ModelGeneralOne,
                                         sort_key=sort_func)
        self.assertEqual(len(models_one), 2)
        self.assertEqual(models_one[0].f_integer, 10)
        self.assertEqual(models_one[0].f_text, '10-second')
        self.assertEqual(models_one[1].f_integer, 20)
        self.assertEqual(models_one[1].f_text, '20-first')
        
    
    #--- allow_multi_update ----------------------------------------------------
    
    def test_multiple_model_update(self):
        model_one_cls = self.ModelGeneralOne
        model_two_cls = self.ModelGeneralTwo
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}, {'f_float', 'f_text'}]
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer'}]
            getters = [{'f_integer'}, {'f_float', 'one_x_1'}]
        
        persister = self.persister
        
        # with simple fields ---------------------------------------------------
        # creating two items with different integer fields but same other fields
        item = ItemGeneralOne(f_integer='100', f_float='200.200',
                              f_boolean=True)
        persister.persist(item)
        item = ItemGeneralOne(f_integer='200', f_float='200.200',
                              f_boolean=False)
        persister.persist(item)
        
        # updating to get 'f_float' and 'f_text' have the same values for
        # different models 
        item = ItemGeneralOne(f_integer='100', f_text='text-1')
        persister.persist(item)
        item = ItemGeneralOne(f_integer='200', f_text='text-1')
        persister.persist(item)
        
        self.assertEqual(len(self.get_all_models(model_one_cls)), 2)
        
        # this must get two models from database
        item = ItemGeneralOne(f_integer='100',
                              f_float='200.200', f_text='text-1')
        ItemGeneralOne.allow_multi_update= True
        items, model_lists = persister.persist(item)
        ItemGeneralOne.allow_multi_update= False
        
        self.assertEqual(len(items), 1)
        self.assertEqual(len(model_lists), 1)
        self.assertEqual(len(model_lists[0]), 2)
        models = model_lists[0]
        models.sort(key=lambda model: not model.f_boolean)
        
        self.assertIs(models[0].f_boolean, True)
        self.assertEqual(models[0].f_integer, 100)
        self.assertEqual(models[0].f_float, 200.2)
        self.assertEqual(models[0].f_text, 'text-1')
        
        self.assertIs(models[1].f_boolean, False)
        self.assertEqual(models[1].f_integer, 100)
        self.assertEqual(models[1].f_float, 200.2)
        self.assertEqual(models[1].f_text, 'text-1')
        
        # still 2 models in database
        self.assertEqual(len(self.get_all_models(model_one_cls)), 2)
        
        # using relations ------------------------------------------------------
        item = ItemGeneralTwo(f_integer='100', f_float='200.200')
        persister.persist(item)
        item = ItemGeneralTwo(f_integer='200', f_float='200.200')
        persister.persist(item)
        
        item_one = ItemGeneralOne(f_integer='300')  # new item one
        item = ItemGeneralTwo(f_integer='100', one_x_1=item_one)
        persister.persist(item)
        item = ItemGeneralTwo(f_integer='200', one_x_1=item_one)
        persister.persist(item)
        
        self.assertEqual(len(self.get_all_models(model_one_cls)), 3)
        self.assertEqual(len(self.get_all_models(model_two_cls)), 2)

        # this must get two models from database
        item = ItemGeneralTwo(f_float='200.200', one_x_1=item_one)
        ItemGeneralTwo.allow_multi_update = True
        item, model_lists = persister.persist(item)
        ItemGeneralTwo.allow_multi_update = False
        
        self.assertEqual(len(items), 1)
        self.assertEqual(len(model_lists), 1)
        self.assertEqual(len(model_lists[0]), 2)
        models = model_lists[0]
        models.sort(key=lambda model: model.f_integer)
        
        self.assertEqual(models[0].f_integer, 100)
        self.assertEqual(models[0].f_float, 200.2)
        
        self.assertEqual(models[1].f_integer, 200)
        self.assertEqual(models[1].f_float, 200.2)
        
        self.assertIs(models[0].one_x_1, models[1].one_x_1)
        self.assertEqual(models[0].one_x_1.f_integer, 300)
    
    
    def test_set_multiple_related(self):
        model_one_cls = self.ModelGeneralOne
        
        class ItemGeneralOne(Item):
            allow_multi_update = True
            model_cls = self.ModelGeneralOne
            creators = [{'f_integer'}]
            getters = [{'f_integer'}, {'f_float', 'f_text'}]
        
        class ItemGeneralTwo(Item):
            allow_multi_update = True
            model_cls = self.ModelGeneralTwo
            creators = [{'f_integer'}]
            getters = [{'f_integer'}, {'f_float', 'one_x_1'}]
        
        persister = self.persister
        
        # creating two items with different integer fields but same other fields
        item = ItemGeneralOne(f_integer='100', f_float='200.200',
                              f_boolean=True)
        persister.persist(item)
        item = ItemGeneralOne(f_integer='200', f_float='200.200',
                              f_boolean=False)
        persister.persist(item)
        
        # updating to get 'f_float' and 'f_text' have the same values for
        # different models 
        item_one = ItemGeneralOne(f_integer='100', f_text='text-1')
        persister.persist(item_one)
        item_one = ItemGeneralOne(f_integer='200', f_text='text-1')
        persister.persist(item_one)
        
        self.assertEqual(len(self.get_all_models(model_one_cls)), 2)
        
        #--- set two models to x-to-many field must work -----------------------
        item_two = ItemGeneralTwo(f_integer='100')
        item_two['one_1_x'].gen(f_float='200.200', f_text='text-1')
        
        items, model_lists = persister.persist(item_two)
        self.assertEqual(len(items), 1)
        self.assertEqual(len(model_lists), 1)
        self.assertEqual(len(model_lists[0]), 1)
        
        # but the model has two related models
        one_1_x = self.get_all_models(model_one_cls)
        self.assertEqual(len(one_1_x), 2)
        
        one_1_x.sort(key=lambda model: model.f_integer)
        self.assertEqual(one_1_x[0].f_integer, 100)
        self.assertEqual(one_1_x[0].f_float, 200.2)
        self.assertEqual(one_1_x[0].f_text, 'text-1')
        self.assertEqual(one_1_x[1].f_integer, 200)
        self.assertEqual(one_1_x[1].f_float, 200.2)
        self.assertEqual(one_1_x[1].f_text, 'text-1')

    
    #--- allow_merge_items -----------------------------------------------------
    
    def test_allow_merge_items(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = ['f_integer']
            getters = ['f_integer']
            allow_merge_items = True
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = ['f_integer']
            getters = ['f_integer']
            allow_merge_items = True
        
        # one
        item_one_1 = ItemGeneralOne(f_integer='1000')
        
        # two
        item_one_1['two_1_x'].gen(f_integer='1001',
                                  f_text='text-1')
        item_two_2 = item_one_1['two_1_x'].gen(f_integer='1001',
                                               f_string='str-1')
        
        # one
        item_two_2['one_x_x'].gen(f_integer='2001',
                                  f_text='text-2')
        item_two_2['one_x_x'].gen(f_integer='2001',
                                  f_string='str-2')
        # not mergable
        item_two_2['one_x_x'].gen(f_integer='9999',
                                  f_string='str-9')
        item_one_1.process()
        
        expect = {
            'item': {
                'two_1_x': {
                    'defaults': {
                        'one_x_1': {
                            'id': 1
                        }
                    },
                    'bulk': [{
                        'item': {
                            'one_x_x': {
                                'defaults': {
                                    'two_x_x': {
                                        'defaults': {},
                                        'bulk': [{
                                            'id': 2
                                        }]
                                    }
                                },
                                'bulk': [{
                                    'item': {
                                        'f_text': 'text-2',
                                        'two_x_x': {
                                            'defaults': {},
                                            'bulk': [{
                                                'id': 2
                                            }]
                                        },
                                        'f_integer': 2001,
                                        'f_string': 'str-2'
                                    }
                                }, {
                                    'item': {
                                        'two_x_x': {
                                            'defaults': {},
                                            'bulk': [{
                                                'id': 2
                                            }]
                                        },
                                        'f_integer': 9999,
                                        'f_string': 'str-9'
                                    }
                                }]
                            },
                            'f_text': 'text-1',
                            'one_x_1': {
                                'id': 1
                            },
                            'f_integer': 1001,
                            'f_string': 'str-1'
                        },
                        'id': 2
                    }]
                },
                'f_integer': 1000
            },
            'id': 1
        }

        self.assertEqual(item_one_1.to_dict(), expect)
        
        # merging at top level
        bulk_one = ItemGeneralOne.Bulk()
        bulk_one.gen(f_integer='1', f_string='str-1')
        bulk_one.gen(f_integer='1', f_text='text-1')
        bulk_one.process()
        
        expect = {
            'bulk': [
                {
                    'item': {
                        'f_integer': 1,
                        'f_string': 'str-1',
                        'f_text': 'text-1'
                    }
                }
            ],
            'defaults': {}
        }
        self.assertEqual(bulk_one.to_dict(), expect)
        
        
    def test_merging_items_point_each_other(self):
        class ItemMergeOne(Item):
            model_cls = self.ModelMergeOne
            creators = ['f_integer']
            getters = ['f_integer']
        
        class ItemMergeTwo(Item):
            model_cls = self.ModelMergeTwo
            creators = ['f_integer']
            getters = ['f_integer']
            allow_merge_items = True
        
        class ItemMergeThree(Item):
            model_cls = self.ModelMergeThree
            creators = ['f_integer']
            getters = ['f_integer']
            allow_merge_items = True
        
        bulk_one = ItemMergeOne.Bulk(two_x_1__three_x_1__f_float=1.0,
                                     two_x_1__f_integer=10)
        for i in range(1, 3):
            item_one = bulk_one.gen(f_integer=i)
            item_one['two_x_1__three_x_1'](f_integer=20)
        
        # items two and items three must be merged, they all the same and can
        # be merged into one item, but they reference each other
        bulk_one.process()
        
        item_one_1, item_one_2 = bulk_one
        
        self.assertEqual(item_one_1['f_integer'], 1)
        self.assertEqual(item_one_2['f_integer'], 2)
        
        self.assertIs(item_one_1['two_x_1'], item_one_1['two_x_1'])
        two_x_1 = item_one_1['two_x_1']
        
        self.assertEqual(two_x_1['f_integer'], 10)
        three_x_1 = two_x_1['three_x_1']
        
        self.assertEqual(three_x_1['f_integer'], 20)
        self.assertEqual(three_x_1['f_float'], 1.0)
        
        
    def test_merge_items_not_the_same(self):
        class ItemMergeOne(Item):
            model_cls = self.ModelMergeOne
            creators = ['f_integer']
            getters = ['f_integer']
        
        class ItemMergeTwo(Item):
            model_cls = self.ModelMergeTwo
            creators = ['f_integer']
            getters = ['f_integer']
            allow_merge_items = True
        
        class ItemMergeThree(Item):
            model_cls = self.ModelMergeThree
            creators = ['f_integer']
            getters = ['f_integer']
            
        
        bulk_one = ItemMergeOne.Bulk(two_x_1__three_x_1__f_float=1.0,
                                     two_x_1__f_integer=10)
        for i in range(1, 3):
            item_one = bulk_one.gen(f_integer=i)
            item_one['two_x_1__three_x_1'](f_integer=20+i)
        
        with self.assertRaises(ItemsNotTheSame):
            bulk_one.process()
    
    
    def test_merge_bulk_items(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            creators = ['f_integer']
            getters = ['f_integer']
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            creators = ['f_integer']
            getters = ['f_integer']
        
        def gen_bulk_one():
            bulk_one = ItemGeneralOne.Bulk()
            bulk_one.gen(f_integer=1,
                         two_x_x=[ItemGeneralTwo(f_integer=10),
                                  ItemGeneralTwo(f_integer=11)])
            bulk_one.gen(f_integer=2,
                         two_x_x=[ItemGeneralTwo(f_integer=11),
                                  ItemGeneralTwo(f_integer=12)])
            return bulk_one
        
        bulk_one = gen_bulk_one()
        with self.assertRaises(MultipleItemsMatch):
            bulk_one.process()
        
        ItemGeneralOne.allow_merge_items = True
        ItemGeneralTwo.allow_merge_items = True
        bulk_one = gen_bulk_one()
        bulk_one[0]['two_x_x'][1]['f_float'] = 1
        bulk_one[1]['two_x_x'][0]['f_float'] = 2
        
        with self.assertRaises(ItemsNotTheSame):
            bulk_one.process()
        
        bulk_one = gen_bulk_one()
        self.assertIsNot(bulk_one[0]['two_x_x'][1], 
                         bulk_one[1]['two_x_x'][0])
        bulk_one.process()
        self.assertEqual(len(bulk_one[0]['two_x_x']), 2)
        self.assertEqual(len(bulk_one[1]['two_x_x']), 2)
        self.assertIs(bulk_one[0]['two_x_x'][1], 
                      bulk_one[1]['two_x_x'][0])
        two_x_x = []
        for item_one in bulk_one:
            for item_two in item_one['two_x_x']:
                if item_two not in two_x_x:
                    two_x_x.append(item_two)
        self.assertEqual(len(two_x_x), 3)


from .test_adapter import TestAdapter
from .test_exceptions import TestExceptions
from .test_item_hooks import TestItemHooks
from .test_resolve_model import TestResolveModel
from .test_model_deleter import TestModelDeleter
from .test_norewrite import TestNorewrite
from .test_persist import TestPersist
from .test_replace import TestReplace
from .test_simple_config import TestSimpleConfig


from django.db import models


# this model must not get into Django migrations as it contains "__" in it's
# name
class ModelInvalidFieldNames(models.Model):
    class Meta:
        db_table = 'model_invalid_field_names'
        
    f__integer = models.IntegerField(null=True)




from save_to_db.exceptions.scope_except import (ItemClsAlreadyScoped,
                                                ScopeIdCannotBeNone)
from save_to_db.exceptions.item_cls_collection import \
    CollectionDoesNotExist, CollectionIdAlreadyInUse
from save_to_db.core.item import Item
from save_to_db.core.scope import Scope
from save_to_db.utils.test_base import TestBase



class TestScope(TestBase):
    
    ModelFieldTypes = None
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        
        class ItemFieldTypes(Item):
            model_cls = cls.ModelFieldTypes
        
        class ItemGeneralOne(Item):
            model_cls = cls.ModelGeneralOne
                
        class ItemGeneralTwo(Item):
            model_cls = cls.ModelGeneralTwo
        
        cls.ItemFieldTypes = ItemFieldTypes
        cls.ItemGeneralOne = ItemGeneralOne
        cls.ItemGeneralTwo = ItemGeneralTwo
    
    
    def test_relation_replacement(self):
        default_date_formats = \
            self.ItemGeneralOne.conversions['date_formats']
            
        new_date_formats = ['%d.%m.%Y']
        self.assertNotEqual(new_date_formats, default_date_formats)
        
        collection_id = 'test_relation_replacement'
        scope = Scope(
            fixes={
                self.ItemGeneralOne: {
                    'conversions': {
                        'date_formats': new_date_formats,
                    }
                }
            },
            collection_id='test_relation_replacement')
        self.assertEqual(len(scope._classes), 2)  # only 2 items scoped
        
        ScopedTypes, ScopedGeneralOne, ScopedGeneralTwo = \
            scope.get(self.ItemFieldTypes,
                      self.ItemGeneralOne,
                      self.ItemGeneralTwo)
        
        # collection_id value must be set
        self.assertEqual(ScopedTypes.get_collection_id(), collection_id)
        self.assertEqual(ScopedGeneralOne.get_collection_id(), collection_id)
        self.assertEqual(ScopedGeneralTwo.get_collection_id(), collection_id)
        
        # scoped by default
        self.assertIsNot(ScopedTypes, self.ItemFieldTypes)
        # scoped directly
        self.assertIsNot(ScopedGeneralOne, self.ItemGeneralOne)
        # scoped through reference
        self.assertIsNot(ScopedGeneralTwo, self.ItemGeneralTwo)
        
        # checking returned items
        expected = {
            ScopedTypes: default_date_formats,
            ScopedGeneralOne: new_date_formats,  # changed
            ScopedGeneralTwo: default_date_formats,
            
            self.ItemFieldTypes: default_date_formats,
            self.ItemGeneralOne: default_date_formats,
            self.ItemGeneralTwo: default_date_formats,
        }
        for item_cls, expected_date_formats in expected.items():
            actual_date_formats = item_cls.conversions['date_formats']
            self.assertEqual(actual_date_formats, expected_date_formats)

        # relations of scoped items fixed
        two_1_1_cls = ScopedGeneralOne.relations['two_1_1']['item_cls']
        self.assertIs(two_1_1_cls, ScopedGeneralTwo)
        self.assertEqual(ScopedGeneralOne()['two_1_1']().get_collection_id(),
                         collection_id)
        
        one_x_x_cls = ScopedGeneralTwo.relations['one_x_x']['item_cls']
        self.assertIs(one_x_x_cls, ScopedGeneralOne)
        self.assertEqual(
            ScopedGeneralTwo()['one_x_x'].gen().get_collection_id(),
            collection_id,
        )
        
        # relations of not scoped items left as they were
        two_1_1_cls = \
            self.ItemGeneralOne.relations['two_1_1']['item_cls']
        self.assertIs(two_1_1_cls, self.ItemGeneralTwo)
        self.assertIsNone(
            self.ItemGeneralOne()['two_1_1']().get_collection_id()
        )
        
        one_x_x_cls = \
            self.ItemGeneralTwo.relations['one_x_x']['item_cls']
        self.assertIs(one_x_x_cls, self.ItemGeneralOne)
        self.assertIsNone(
            self.ItemGeneralTwo()['one_x_x'].gen().get_collection_id()
        )
    
    
    def test_none_and_true_item_cls(self):
        scope = Scope(
            fixes={
                self.ItemGeneralOne: {
                    'update_only_mode': True,
                },
                None: {
                    'allow_merge_items': True
                },
                True: {
                    'allow_multi_update': True,
                }
            },
            collection_id='test_none_and_true_item_cls')
        self.assertEqual(len(scope._classes), 3)  # all items scoped
        
        defaults = {
            'update_only_mode': False,
            'allow_merge_items': False,
            'allow_multi_update': False,
        }
        for attr, value in defaults.items():
            for cls in [self.ItemFieldTypes,
                        self.ItemGeneralOne,
                        self.ItemGeneralTwo,]:
                self.assertEqual(getattr(cls, attr), value)
        
        ScopedTypes, ScopedGeneralOne, ScopedGeneralTwo = \
            scope.get(self.ItemFieldTypes,
                      self.ItemGeneralOne,
                      self.ItemGeneralTwo)
        
        after_scoping = {
            ScopedTypes: {
                'update_only_mode': False,
                'allow_merge_items': True,  # by `None`
                'allow_multi_update': True,  # by `True`
            },
            ScopedGeneralOne: {
                'update_only_mode': True,  # by class
                'allow_merge_items': False,
                'allow_multi_update': True,  # by `True`
            },
            ScopedGeneralTwo: {
                'update_only_mode': False,
                'allow_merge_items': True,  # by `None`
                'allow_multi_update': True,  # by `True`
            },
        }
        for scoped_cls, attrs in after_scoping.items():
            for attr, value in attrs.items():
                self.assertEqual(
                    getattr(scoped_cls, attr), value,
                    'scope: {}, cls: {}, attr: {}, val: {}'.format(
                        scoped_cls.get_collection_id(),
                        scoped_cls.__name__, attr, getattr(scoped_cls, attr)))
    
    
    def test_merging_defaults_and_conversions(self):
        self.item_cls_manager.clear()
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
                
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
        
        ItemGeneralOne.complete_setup()
        ItemGeneralTwo.complete_setup()
        
        # default_decimal_separator
        default_decimal_separator = \
            ItemGeneralOne.conversions['decimal_separator']
        # new_descimal_separator
        new_descimal_separator = '!'
        self.assertNotEqual(
            new_descimal_separator, default_decimal_separator)
        
        default_date_formats = \
            self.ItemGeneralOne.conversions['date_formats']
        new_date_formats = ['%d.%m.%Y']
        self.assertNotEqual(new_date_formats, default_date_formats)
        
        expected = {
            ItemGeneralOne: {
                'conversions': {
                    'decimal_separator': default_decimal_separator,
                    'date_formats': default_date_formats,
                },
                'defaults': {}
            },
            ItemGeneralTwo: {
                'conversions': {
                    'decimal_separator': default_decimal_separator,
                    'date_formats': default_date_formats,
                },
                'defaults': {}
            }
        }
        for item_cls, attrs in expected.items():
            for attr, dict_value in attrs.items():
                if not dict_value:
                    self.assertEqual(getattr(item_cls, attr), dict_value)
                    continue
                for key, value in dict_value.items():
                    self.assertEqual(
                        getattr(item_cls, attr)[key], value)
        
        # scoping
        scope = Scope(
            fixes={
                ItemGeneralOne: {
                    'defaults': {
                        'f_integer': 100,
                    },
                },
                ItemGeneralTwo: {
                    'conversions': {
                        'decimal_separator': '!',
                    },
                },
                True: {
                    'conversions': {
                        'decimal_separator': new_descimal_separator,
                        'date_formats': new_date_formats,
                    },
                    'defaults': {
                        'f_integer': 200,
                        'f_text': 'text-200',
                    },
                }
            },
            collection_id='test_merging_defaults_and_conversions')
        
        ScopedItemOne, ScopedItemTwo = scope.get(
            ItemGeneralOne, ItemGeneralTwo)
        expected = {
            ScopedItemOne: {
                'conversions': {
                    'decimal_separator': new_descimal_separator,
                    'date_formats': new_date_formats,
                },
                'defaults': {
                    'f_integer': 100,  # not overwritten
                    'f_text': 'text-200',
                }
            },
            ScopedItemTwo: {
                'conversions': {
                    'decimal_separator': '!',  # not overwritten
                    'date_formats': new_date_formats,
                },
                'defaults': {
                    'f_integer': 200,
                    'f_text': 'text-200',
                }
            }
        }
        for item_cls, attrs in expected.items():
            for attr, dict_value in attrs.items():
                if not dict_value:
                    self.assertEqual(getattr(item_cls, attr), dict_value)
                for key, value in dict_value.items():
                    self.assertEqual(
                        getattr(item_cls, attr)[key], value)
        
    
    
    def test_gen_non_scoped_cls(self):
        self.item_cls_manager.clear()
        
        class ItemFieldTypes(Item):
            model_cls = self.ModelFieldTypes
        
        scope = Scope(
            fixes={
                ItemFieldTypes: {
                    'conversions': {
                        'decimal_separator': '!',
                    }
                },
                True: {
                    'conversions': {
                        'date_formats': ['TRUE: %Y-%m-%d'],
                    }
                },
                None: {
                    'conversions': {
                        'datetime_formats': ['NONE: %Y-%m-%d %H:%M:%S'],
                    }
                }
            },
            collection_id='test_gen_non_scoped_cls')
        self.assertEqual(len(scope._classes), 1)  # only 1 items scoped
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
                
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
        
        ScopedItemOne = scope[ItemGeneralOne]
        self.assertEqual(len(scope._classes), 3)  # rest of the items scoped
        
        ScopedItemTwo = ScopedItemOne.relations['two_x_x']['item_cls']
        
        for item_cls in (ScopedItemOne, ScopedItemTwo):
            self.assertEqual(item_cls.get_collection_id(),
                             'test_gen_non_scoped_cls')
            conversions_two = item_cls.conversions
            
            self.assertEqual(conversions_two['date_formats'],
                             ['TRUE: %Y-%m-%d'])
            self.assertEqual(conversions_two['datetime_formats'],
                             ['NONE: %Y-%m-%d %H:%M:%S'])
        
    
    def test_item_already_scoped_exception(self):
        scope = Scope(
            fixes={
                self.ItemFieldTypes: {
                    'conversions': {
                        'date_formats': '%m/%d/%Y',
                    }
                },
            },
            collection_id='test_item_already_scoped_exception_1')
        ScopedTypes = scope.get(self.ItemFieldTypes)[0]
        
        with self.assertRaises(ItemClsAlreadyScoped):
            scope = Scope(
                fixes={
                    ScopedTypes: {
                        'conversions': {
                            'date_formats': '%d-%m-%Y',
                        }
                    },
                },
                collection_id='test_item_already_scoped_exception_2')
    
    
    def test_item_cls_autogeneration(self):
        self.item_cls_manager.clear()
        self.item_cls_manager.autogenerate = True
        
        default_date_formats = \
            self.ItemGeneralOne.conversions['date_formats']
        new_date_formats = ['%m/%d/%Y']
        self.assertNotEqual(new_date_formats, default_date_formats)
        
        class ItemFieldTypes(Item):
            model_cls = self.ModelFieldTypes
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
        
        scope = Scope(
            fixes={
                ItemGeneralOne: {
                    'conversions': {
                        'date_formats': new_date_formats,
                    }
                },
            },
            collection_id='test_item_cls_autogeneration')
        
        # 2 items including related (ItemFieldTypes discarded)
        self.assertEqual(len(scope._classes), 2)
        
        ItemGeneralTwo = ItemGeneralOne.relations['two_x_x']['item_cls']
        
        self.assertTrue(ItemGeneralTwo.model_cls, self.ModelGeneralTwo)
        
        self.assertIn(ItemGeneralOne, scope._classes)
        self.assertIn(ItemGeneralTwo, scope._classes)
        
        self.assertFalse(ItemFieldTypes.metadata['autogenerated_item_cls'])
        self.assertFalse(ItemGeneralOne.metadata['autogenerated_item_cls'])
        self.assertTrue(ItemGeneralTwo.metadata['autogenerated_item_cls'])
        
        self.assertEqual(ItemFieldTypes.conversions['date_formats'],
                         default_date_formats)
        self.assertEqual(ItemGeneralOne.conversions['date_formats'],
                         default_date_formats)
        self.assertEqual(ItemGeneralTwo.conversions['date_formats'],
                         default_date_formats)
        
        
        self.assertEqual(scope[ItemFieldTypes].conversions['date_formats'],
                         default_date_formats)
        self.assertEqual(scope[ItemGeneralOne].conversions['date_formats'],
                         new_date_formats)
        self.assertEqual(scope[ItemGeneralTwo].conversions['date_formats'],
                         default_date_formats)
        
    
    def test_scope_id_cannot_be_none(self):
        with self.assertRaises(ScopeIdCannotBeNone):
            Scope(fixes={}, collection_id=None)
            
    
    def test_scope_id_already_in_use(self):
        Scope(fixes={}, collection_id='test_scope_id_already_in_use')
        
        with self.assertRaises(CollectionIdAlreadyInUse):
            Scope(fixes={}, collection_id='test_scope_id_already_in_use')
        
    
    def test_scope_does_not_exist(self):
        with self.assertRaises(CollectionDoesNotExist):
            Scope.get_collection_by_id(
                collection_id='test_scope_does_not_exist')
    
    
    def test_deleter_recreation(self):
        self.item_cls_manager.clear()
        self.item_cls_manager.autogenerate = True
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            deleter_selectors = ['f_integer']
        ItemGeneralOne.complete_setup()
         
        scope = Scope(
            fixes={
                ItemGeneralOne: {},
            },
            collection_id='test_deleter_recreation')
         
        self.assertIsNotNone(ItemGeneralOne.metadata['model_deleter'])
        ScopedItemOne = scope[ItemGeneralOne]
        ScopedItemOne.complete_setup()
        self.assertIsNotNone(ScopedItemOne.metadata['model_deleter'])
         
        self.assertIsNot(ScopedItemOne.metadata['model_deleter'],
                         ItemGeneralOne.metadata['model_deleter'])
    
    
    def test_unref_recreation(self):
        self.item_cls_manager.clear()
        self.item_cls_manager.autogenerate = True
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            unref_x_to_many = {
                'two_x_x': {'f_string'},
            }
        ItemGeneralOne.complete_setup()
        
        scope = Scope(
            fixes={
                ItemGeneralOne: {},
            },
            collection_id='test_unref_recreation')
        ScopedItemOne = scope[ItemGeneralOne]
        ScopedItemOne.complete_setup()
        
        self.assertTrue(ItemGeneralOne.metadata['model_unrefs'])
        self.assertTrue(ScopedItemOne.metadata['model_unrefs'])
        
        self.assertIn('two_x_x', ItemGeneralOne.metadata['model_unrefs'])
        self.assertIn('two_x_x', ScopedItemOne.metadata['model_unrefs'])
        
        self.assertIsNot(ScopedItemOne.metadata['model_unrefs']['two_x_x'],
                         ItemGeneralOne.metadata['model_unrefs']['two_x_x'])
        


from django.db import models



class ModelFieldTypes(models.Model):
    
    class Meta:
        db_table = 'model_field_types'
        
    binary_1 = models.BinaryField(db_column='new_binary_1')
    
    string_1 = models.CharField(db_column='new_string_1', max_length=255)
    
    text_1 = models.TextField(db_column='new_text_1')
    
    integer_1 = models.IntegerField(db_column='new_integer_1')
    integer_2 = models.BigIntegerField(db_column='new_integer_2')
    integer_3 = models.PositiveIntegerField(db_column='new_integer_3')
    integer_4 = models.PositiveSmallIntegerField(db_column='new_integer_4')
    integer_5 = models.SmallIntegerField(db_column='new_integer_5')
    
    boolean_1 = models.BooleanField(db_column='new_boolean_1')
    boolean_2 = models.NullBooleanField(db_column='new_boolean_2')
    
    float_1 = models.FloatField(db_column='new_float_1')
    
    decimal_1 = models.DecimalField(db_column='new_decimal_2',
                                    max_digits=12, decimal_places=3)
    
    date_1 = models.DateField(db_column='new_date_1')
    
    time_1 = models.TimeField(db_column='new_time_1')
    
    datetime_1 = models.DateTimeField(db_column='new_datetime_1')



import pprint


class DictWrapper(dict):
    """ This class is used as a temporary wrapper around
    :py:class:`~save_to_db.core.item_base.ItemBase` instance for compatibility
    with Scrapy project.
    
    Scrapy is an open source and collaborative framework for extracting the data
    you need from websites. In a fast, simple, yet extensible way.
    [#DictWrapper_1]_
    
    .. [#DictWrapper_1] Citation from https://scrapy.org site.

    When parsing a page with Scrapy you cannot yield instances of arbitrary
    classes, but you can yield an instance of a `dict` class which is treated
    as an instance of a Scrapy item and properly sent to item pipelines.
    
    Using :py:meth:`~save_to_db.core.item_base.ItemBase.to_dict` and then
    :py:meth:`~save_to_db.core.item_base.ItemBase.load_dict` of
    :py:class:`~save_to_db.core.item_base.ItemBase` instance is expensive, as it
    properly transforms an item into a `dict` and then `dict` to item. Using
    :py:meth:`~save_to_db.core.item_base.ItemBase.dict_wrapper` method of
    :py:class:`~save_to_db.core.item_base.ItemBase` will just wrap an item in
    a :py:class:`~.DictWrapper` instance (subclass of `dict` class),
    later you can get the wrapped item from the wrapper in a Scrapy pipeline. 
    
    Instances of Scrapy items and of this library items are not completely
    compatible, so you need to use different item pipelines for this library
    items, the pipelines must accept and return instances of
    :py:class:`~.DictWrapper` if you want to use more then one pipeline.
    
    Here is an example of a Scrapy pipeline that saves items to a database:
    
        .. code-block:: Python
        
            from django.db import transaction
            from save_to_db import Persister
            from save_to_db.adapters import DjangoAdapter

            persister = Persister(DjangoAdapter, adapter_settings={})

            class DbPipeline(object):
            
                def process_item(self, item, spider):
                    stdb_item = item.get_item()
                    
                    with transaction.atomic():
                        persister.persist(stdb_item)
                    
                    return item  # returning item wrapper
                    
        .. note:
            Printing item wrapper to a console will output the wrapped item
            as a dictionary.  
    
    
    :param item: An instance of :py:class:`~save_to_db.core.item_base.ItemBase`
        to be wrapped in a dictionary.
    """
    
    def __init__(self, item):
        self['item'] = item
    
    def get_item(self):
        """ Returns an originally wrapped item instance.
        
        :returns: Instance of
            :py:class:`~save_to_db.core.item_base.ItemBase` class.
        """
        return self['item']
    
    def __str__(self):
        return pprint.pformat(self.get_item().to_dict())


import copy
from inspect import isclass
from save_to_db.exceptions.scope_except import (ItemClsAlreadyScoped,
                                                ScopeIdCannotBeNone)
from save_to_db.exceptions.item_cls_collection import CollectionIdAlreadyInUse
from .item_base import ItemBase
from .item_cls_manager import item_cls_manager
from .utils.item_cls_collection import ItemClsCollection


class Scope(ItemClsCollection):
    """ Class for scoping :py:class:`~.item.Item` classes.
    
    :param fixes: a dictionary whose keys are item classes (or `None`
        for all classes not in the dictionary) and values are class
        attributes to be replaced.
    :param collection_id: Scope ID, can be value of any type.
    
        .. seealso::
            :py:meth:`~.item_base.ItemBase.get_collection_id` method of
            :py:class:`~.item_base.ItemBase` class.
    
    Example usage:
        
        .. code:: Python
        
            class TestItem(Item):
                model_cls = SomeModel
            
            scope = Scope({
                TestItem: {
                    'conversions': {
                        'date_formats': '%m/%d/%Y',
                    },
                },
                # for item classes not listed
                None: {
                    'conversions': {
                        'date_formats': '%d.%m.%Y',
                    },
                },
                # for all item classes
                True: 'conversions': {
                    'datetime_formats': '%Y-%m-%d %H:%M:%S.%f',
                }, 
            })
            
            ScopedTestItem = scope[TestItem]  # or `scope.get(TestItem)`
    
    When an item is scoped other items that use the original item in
    relations are also scoped and their relation data fixed.
    """
    
    def __init__(self, fixes, collection_id):
        if collection_id is None:
            raise ScopeIdCannotBeNone()
        if collection_id in Scope.collection_classes:
            raise CollectionIdAlreadyInUse(collection_id)
        super().__init__(collection_id=collection_id)
        
        self.none_cls_fixes = {}
        self._classes = {}
        item_cls_manager.autocomplete_item_classes()
        
        # injecting from `True`
        if True in fixes:
            for item_cls, item_fixes in fixes.items():
                if item_cls is True:
                    continue
                item_fixes = self.__merge_fixes(item_fixes, fixes[True])
                
            if None not in fixes:
                fixes[None] = fixes[True]
        
        # fixing classes
        for item_cls, item_fixes in fixes.items():
            if item_cls is None or item_cls is True:
                continue
            self.__prepare_scoped_item_cls(item_cls, item_fixes)
        
        # `None` class
        if None in fixes:
            self.none_cls_fixes = copy.deepcopy(fixes[None])
            for item_cls in item_cls_manager.get_all_item_classes():
                if item_cls in self._classes:
                    continue
                self.__prepare_scoped_item_cls(item_cls, fixes[None])
        
        self.__fix_relations()
        
    
    def __fix_relations(self, item_fixes=None):
        if item_fixes is None:
            item_fixes = {}
        # fixing relations
        do_continue = True
        while do_continue:
            do_continue = False
            #--- first scoping items that need to be scoped ---
            for item_cls in item_cls_manager.get_all_item_classes():
                if item_cls in self._classes:
                    continue
                
                item_cls.complete_setup()
                for rel_data in item_cls.relations.values():
                    if rel_data['item_cls'] in self._classes:
                        self.__prepare_scoped_item_cls(item_cls,
                                                       item_fixes=item_fixes)
                        break
            
            #--- updating relations if needed ---
            for scoped_item_cls in self._classes.values():
                relations = scoped_item_cls.relations
                for key, rel_data in relations.items():
                    rel_item_cls = rel_data['item_cls']
                    if rel_item_cls not in self._classes:
                        continue
                    
                    relations[key]['item_cls'] = self._classes[rel_item_cls]
                    do_continue = True
    
    
    def __merge_fixes(self, extened, merged):
        merged_is_item = isclass(merged) and issubclass(merged, ItemBase)
        if merged_is_item:
            keys = list(extened.keys())
        else:
            keys = list(merged.keys())
        
        for key in keys:
            if merged_is_item:
                merging_value = getattr(merged, key)
            else:
                merging_value = merged[key]
            merging_value = copy.deepcopy(merging_value)
            
            if key not in extened:
                extened[key] = merging_value
                continue
            
            if key in ['defaults', 'conversions']:
                merging_value.update(extened[key])
                extened[key] = merging_value 

        return extened
        
    
    def __prepare_scoped_item_cls(self, item_cls, item_fixes):
        if item_cls.metadata['collection_id'] != None:
            raise ItemClsAlreadyScoped(item_cls)

        item_cls.complete_setup()
        item_fixes = copy.deepcopy(item_fixes)
        item_fixes['relations'] = copy.deepcopy(item_cls.relations)
        item_fixes['metadata'] = {}
        for key, value in item_cls.metadata.items():
            if key not in ('collection_id', 'model_deleter', 'model_unrefs',
                           'setup_completed'):
                item_fixes['metadata'][key] = value
        item_fixes['metadata']['collection_id'] = self.collection_id
        
        class_name = item_cls.__name__
        scoped_item_cls = type('Scoped{}'.format(class_name), (item_cls,),
                               self.__merge_fixes(item_fixes, item_cls))
        self._classes[item_cls] = scoped_item_cls
        super().add(scoped_item_cls)
    
    
    def __getitem__(self, item_cls):
        if item_cls not in self._classes:
            item_cls.complete_setup()
            item_cls_manager.autocomplete_item_classes()
            self.__prepare_scoped_item_cls(item_cls, self.none_cls_fixes)
            self.__fix_relations(self.none_cls_fixes)
        return self._classes[item_cls]
    
    
    def get(self, *item_classes):
        """ Excepts non-scoped items and returns corresponding scoped items or
        original items for items not present in the scope.
        
        :param item_classes: Items for which scoped item versions going to be
            returned.
        :returns: List of scoped items. If scope does not have corresponding
            scoped version of an item class, original class is used.
        """
        return [self[item_cls] for item_cls in item_classes]
    
    #--- managing scopes -------------------------------------------------------
    
    @classmethod
    def delete_all_scopes(cls):
        """ Removes all scopes. """
        for key in list(cls.collection_classes.keys()):
            if key is not None:
                del cls.collection_classes[key]



import datetime
from decimal import Decimal

from io import BytesIO
from save_to_db.core.item import Item
from save_to_db.utils.test_base import TestBase



class TestItemDump(TestBase):

    ModelGeneralOne = None
    ModelGeneralTwo = None

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        
        class ItemFieldTypes(Item):
            model_cls = cls.ModelFieldTypes
        cls.ItemFieldTypes = ItemFieldTypes
        
        class ItemGeneralOne(Item):
            model_cls = cls.ModelGeneralOne
        cls.ItemGeneralOne = ItemGeneralOne
        
        class ItemGeneralTwo(Item):
            model_cls = cls.ModelGeneralTwo
        cls.ItemGeneralTwo = ItemGeneralTwo
    
    
    def check_dumps(self, item, expected):
        # converting to bytes
        item_as_bytes = self.persister.dumps(item)
        self.assertIsInstance(item_as_bytes, bytes)
        
        item = self.persister.loads(item_as_bytes)    
        self.assertEqual(item.to_dict(), expected)
        
        # saving to a file-like object        
        storage = BytesIO()
        self.persister.dump(item, storage)
    
        storage.seek(0)
        item = self.persister.load(storage)
        self.assertEqual(item.to_dict(), expected)
    
    
    def test_dump_types(self):
        item = self.ItemGeneralOne(f_binary='binary data',
                                   f_boolean='true',
                                   f_string='string data',
                                   f_text='text_data',
                                   f_integer='10',
                                   f_float='10.10',
                                   f_date='2000-10-30',
                                   f_time='20:30:40',
                                   f_datetime='2000-10-30 20:30:40')
        expected = {
            'item': {
                'f_binary': b'binary data',
                'f_boolean': True,
                'f_date': datetime.date(2000, 10, 30),
                'f_datetime': datetime.datetime(2000, 10, 30, 20, 30, 40),
                'f_float': 10.1,
                'f_integer': 10,
                'f_string': 'string data',
                'f_text': 'text_data',
                'f_time': datetime.time(20, 30, 40),
            }
        }
        self.check_dumps(item, expected)


    def test_dump_single_item_with_relations(self):
        item = self.ItemGeneralOne(f_integer='1')
        item['two_1_1__f_integer'] = '11'
        item['two_x_x'].gen(f_float='99.1')
        item['two_x_x'].gen(f_float='99.2', one_x_1__f_integer='192') 
        item['two_x_x__f_float'] = '1000.1'
        
        expected = {
            'id': 1,
            'item': {
                'two_x_x': {
                    'defaults': {
                        'one_x_x': {
                            'defaults': {},
                            'bulk': [{
                                'id': 1
                            }]
                        },
                        'f_float': 1000.1
                    },
                    'bulk': [{
                        'item': {
                            'one_x_x': {
                                'defaults': {},
                                'bulk': [{
                                    'id': 1
                                }]
                            },
                            'f_float': 99.1
                        }
                    }, {
                        'id': 3,
                        'item': {
                            'one_x_1': {
                                'id': 2,
                                'item': {
                                    'two_1_x': {
                                        'defaults': {
                                            'one_x_1': {
                                                'id': 2
                                            }
                                        },
                                        'bulk': [{
                                            'id': 3
                                        }]
                                    },
                                    'f_integer': 192
                                }
                            },
                            'f_float': 99.2,
                            'one_x_x': {
                                'defaults': {},
                                'bulk': [{
                                    'id': 1
                                }]
                            }
                        }
                    }]
                },
                'f_integer': 1,
                'two_1_1': {
                    'item': {
                        'one_1_1': {
                            'id': 1
                        },
                        'f_integer': 11
                    }
                }
            }
        }
        
        self.check_dumps(item, expected)
        
        
    def test_dump_bulk_item_with_relations(self):
        bulk = self.ItemGeneralOne.Bulk()
        bulk.gen(f_float='1.1')
        bulk.gen(f_float='1.2', two_x_1__f_integer='2') 
        bulk['two_x_x__f_float'] = '1000.1'
        bulk['two_x_1'] = self.ItemGeneralTwo(f_boolean='true')

        expected = {
            'bulk': [{
                'id': 2
            }, {
                'item': {
                    'two_x_1': {
                        'item': {
                            'f_integer': 2,
                            'one_1_x': {
                                'bulk': [{
                                    'id': 4
                                }],
                                'defaults': {
                                    'two_x_1': {
                                        'id': 3
                                    }
                                }
                            }
                        },
                        'id': 3
                    },
                    'f_float': 1.2,
                    'two_x_x': {
                        'bulk': [],
                        'defaults': {
                            'f_float': 1000.1
                        }
                    }
                },
                'id': 4
            }],
            'defaults': {
                'two_x_x__f_float': 1000.1,
                'two_x_1': {
                    'item': {
                        'f_boolean': True,
                        'one_1_x': {
                            'bulk': [{
                                'item': {
                                    'two_x_1': {
                                        'id': 1
                                    },
                                    'f_float': 1.1,
                                    'two_x_x': {
                                        'bulk': [],
                                        'defaults': {
                                            'f_float': 1000.1
                                        }
                                    }
                                },
                                'id': 2
                            }],
                            'defaults': {
                                'two_x_1': {
                                    'id': 1
                                }
                            }
                        }
                    },
                    'id': 1
                }
            }
        }

        self.check_dumps(bulk, expected)
    
    
    def test_multiple_items_dump(self):
        storage = BytesIO()
        
        self.persister.dump(self.ItemGeneralOne(f_integer='1'), storage)
        self.persister.dump(self.ItemGeneralOne(f_integer='2'), storage)
        self.persister.dump(self.ItemGeneralOne(f_integer='3'), storage)
        
        storage.seek(0)
        
        dict_wrapper = self.persister.load(storage).to_dict()
        self.assertEqual(dict_wrapper, {'item': {'f_integer': 1}})
        
        dict_wrapper = self.persister.load(storage).to_dict()
        self.assertEqual(dict_wrapper, {'item': {'f_integer': 2}})
        
        dict_wrapper = self.persister.load(storage).to_dict()
        self.assertEqual(dict_wrapper, {'item': {'f_integer': 3}})
        
        # no more items
        dict_wrapper = self.persister.load(storage)
        self.assertIsNone(dict_wrapper)


    def test_arbitrary_id_position(self):
        # normal item
        item_one = self.ItemGeneralOne(f_integer='1')
        child_item = item_one['child_1_x'].gen(f_integer='2')
        item_one['child_x_x'].add(child_item)
        item_one.process()
        expect = {
            'id': 1,
            'item': {
                'child_1_x': {
                    'bulk': [
                        {
                            'id': 2,
                            'item': {
                                'f_integer': 2,
                                'parent_x_1': {
                                    'id': 1
                                },
                                'parent_x_x': {
                                    'bulk': [
                                        {
                                            'id': 1
                                        }
                                    ],
                                    'defaults': {}
                                }
                            }
                        }
                    ],
                    'defaults': {}
                },
                'child_x_x': {
                    'bulk': [
                        {
                            'id': 2
                        }
                    ],
                    'defaults': {}
                },
                'f_integer': 1
            }
        }
        loaded_item = self.ItemGeneralOne().load_dict(expect)
        self.assertEqual(loaded_item.to_dict(), expect)
        
        # moving parent data
        item_as_dict = {
            'id': 1,
            'item': {
                'child_1_x': {
                    'bulk': [
                        {
                            'id': 2
                        }
                    ],
                    'defaults': {}
                },
                'child_x_x': {
                    'bulk': [
                        {
                            'id': 2,
                            'item': {
                                'f_integer': 2,
                                'parent_x_1': {
                                    'id': 1
                                },
                                'parent_x_x': {
                                    'bulk': [
                                        {
                                            'id': 1
                                        }
                                    ],
                                    'defaults': {}
                                }
                            }
                        }
                    ],
                    'defaults': {}
                },
                'f_integer': 1
            }
        }
        loaded_item = self.ItemGeneralOne().load_dict(item_as_dict)
        self.assertEqual(loaded_item.to_dict(), expect)
        
        # id in default
        item_one = self.ItemGeneralOne.Bulk()
        item_one.gen(f_integer='1')
        other_item = item_one.gen(f_integer='2')
        item_one['parent_x_1'] = other_item
        item_one.process()
        expect = {
            'bulk': [
                {
                    'id': 2,
                    'item': {
                        'f_integer': 1,
                        'parent_x_1': {
                            'item': {
                                'child_1_x': {
                                    'bulk': [
                                        {
                                            'id': 2
                                        }
                                    ],
                                    'defaults': {}
                                },
                                'f_integer': 2
                            }
                        }
                    }
                },
                {
                    'id': 1
                }
            ],
            'defaults': {
                'parent_x_1': {
                    'id': 1,
                    'item': {
                        'f_integer': 2,
                        'parent_x_1': {
                            'item': {
                                'child_1_x': {
                                    'bulk': [
                                        {
                                            'id': 1
                                        }
                                    ],
                                    'defaults': {}
                                },
                                'f_integer': 2
                            }
                        }
                    }
                }
            }
        }
        loaded_item = self.ItemGeneralOne.Bulk().load_dict(expect)
        self.assertEqual(loaded_item.to_dict(), expect)
        
        # moving other item data
        item_as_dict = {
            'bulk': [
                {
                    'id': 2,
                    'item': {
                        'f_integer': 1,
                        'parent_x_1': {
                            'item': {
                                'child_1_x': {
                                    'bulk': [
                                        {
                                            'id': 2
                                        }
                                    ],
                                    'defaults': {}
                                },
                                'f_integer': 2
                            }
                        }
                    }
                },
                {
                    'id': 1,
                    'item': {
                        'f_integer': 2,
                        'parent_x_1': {
                            'item': {
                                'child_1_x': {
                                    'bulk': [
                                        {
                                            'id': 1
                                        }
                                    ],
                                    'defaults': {}
                                },
                                'f_integer': 2
                            }
                        }
                    }
                }
            ],
            'defaults': {
                'parent_x_1': {
                    'id': 1,
                  }
            }
        }
        loaded_item = self.ItemGeneralOne.Bulk().load_dict(item_as_dict)
        self.assertEqual(loaded_item.to_dict(), expect)
    
    
    def test_dump_instance_config(self):
        self.item_cls_manager.clear()
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            update_only_mode = True
        
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
            get_only_mode = True
            
        item_one_bulk = ItemGeneralOne.Bulk()
        item_one_bulk.gen(f_integer=1)
        
        item_one = item_one_bulk.gen(f_integer=2)
        item_one.get_only_mode = True
        item_one.update_only_mode = False
        
        
        item_one['two_1_x'].gen(f_integer=201)
        
        item_two = item_one['two_1_x'].gen(f_integer=202)
        item_two.get_only_mode = False
        item_two.update_only_mode = True
        
        expected = {
            'bulk': [
                {
                    'item': {
                        'f_integer': 1
                    }
                },
                {
                    'id': 1,
                    'item': {
                        'f_integer': 2,
                        'two_1_x': {
                            'bulk': [
                                {
                                    'item': {
                                        'f_integer': 201,
                                        'one_x_1': {
                                            'id': 1
                                        }
                                    }
                                },
                                {
                                    'item': {
                                        'f_integer': 202,
                                        'one_x_1': {
                                            'id': 1
                                        }
                                    },
                                    # overwritten values
                                    'get_only_mode': False,
                                    'update_only_mode': True
                                }
                            ],
                            'defaults': {
                                'one_x_1': {
                                    'id': 1
                                }
                            }
                        }
                    },
                    # overwritten values
                    'get_only_mode': True,
                    'update_only_mode': False
                }
            ],
            'defaults': {}
        }

        self.check_dumps(item_one_bulk, expected)


    def test_revert_and_to_dict(self):
        self.item_cls_manager.clear()
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            conversions = {
                'date_formats': '%d.%m.%Y',
                'time_formats': '%H-%M-%S',
                'datetime_formats': '%d.%m.%Y %H-%M-%S',
                'decimal_separator': '!',
            }
         
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
        
        # from single item -----------------------------------------------------
        one = ItemGeneralOne(f_integer='10',
                             f_decimal='10!10',
                             f_date='20.10.2010',
                             f_time='10-20-30',
                             f_datetime='20.10.2010 10-20-30')
        one['two_1_1'] = ItemGeneralTwo(f_integer='20',
                                        f_decimal='20.20',
                                        f_date='2010-10-20',
                                        f_time='10:20:30',
                                        f_datetime='2010-10-20 10:20:30')
        for i in range(3, 5):
            one['two_1_x'].gen(f_integer=i*10,
                               f_decimal='{}.{}'.format(i*10, i*10),
                               f_date='2010-10-20',
                               f_time='10:20:30',
                               f_datetime='2010-10-20 10:20:30')
        
        one.process()
        
        expect_processed = {
            'id': 1,
            'item': {
                'f_date': datetime.date(2010, 10, 20),
                'f_datetime': datetime.datetime(2010, 10, 20, 10, 20, 30),
                'f_decimal': Decimal('10.10'),
                'f_integer': 10,
                'f_time': datetime.time(10, 20, 30),
                'two_1_1': {
                    'item': {
                    'f_date': datetime.date(2010, 10, 20),
                       'f_datetime': datetime.datetime(2010, 10, 20,
                                                       10, 20, 30),
                       'f_decimal': Decimal('20.20'),
                       'f_integer': 20,
                       'f_time': datetime.time(10, 20, 30),
                       'one_1_1': {
                           'id': 1
                        }
                    }
                },
                'two_1_x': {
                    'bulk': [
                        {
                            'item': {
                                'f_date': datetime.date(2010, 10, 20),
                                'f_datetime': datetime.datetime(2010, 10, 20,
                                                                10, 20, 30),
                                'f_decimal': Decimal('30.30'),
                                'f_integer': 30,
                                'f_time': datetime.time(10, 20, 30),
                                'one_x_1': {
                                    'id': 1
                                }
                            }
                        },
                        {
                            'item': {
                                'f_date': datetime.date(2010, 10, 20),
                                'f_datetime': datetime.datetime(2010, 10, 20,
                                                                10, 20, 30),
                                'f_decimal': Decimal('40.40'),
                                'f_integer': 40,
                                'f_time': datetime.time(10, 20, 30),
                                'one_x_1': {
                                    'id': 1
                                }
                            }
                        }
                    ],
                    'defaults': {
                        'one_x_1': {
                            'id': 1
                        }
                    }
                }
            }
        }

        self.assertEqual(one.to_dict(), expect_processed)
        
        expect_reverted = {
            'id': 1,
            'item': {
                'f_date': '20.10.2010',
                'f_datetime': '20.10.2010 10-20-30',
                'f_decimal': '10!10',
                'f_integer': 10,
                'f_time': '10-20-30',
                'two_1_1': {
                    'item': {
                        'f_date': '2010-10-20',
                        'f_datetime': '2010-10-20 10:20:30',
                        'f_decimal': '20.20',
                        'f_integer': 20,
                        'f_time': '10:20:30',
                        'one_1_1': {
                            'id': 1
                        }
                    }
                },
                'two_1_x': {
                    'bulk': [
                        {
                            'item': {
                                'f_date': '2010-10-20',
                                'f_datetime': '2010-10-20 10:20:30',
                                'f_decimal': '30.30',
                                'f_integer': 30,
                                'f_time': '10:20:30',
                                'one_x_1': {
                                    'id': 1
                                }
                            }
                        },
                        {
                            'item': {
                                'f_date': '2010-10-20',
                                'f_datetime': '2010-10-20 10:20:30',
                                'f_decimal': '40.40',
                                'f_integer': 40,
                                'f_time': '10:20:30',
                                'one_x_1': {
                                    'id': 1
                                }
                            }
                        }
                    ],
                    'defaults': {
                        'one_x_1': {
                            'id': 1
                        }
                    }
                }
            }
        }

        self.assertEqual(one.to_dict(revert=True), expect_reverted)
        # original not changed
        self.assertEqual(one.to_dict(), expect_processed)
        
        # now all reverted
        one.revert()
        self.assertEqual(one.to_dict(), expect_reverted)
        self.assertEqual(one.to_dict(revert=True), expect_reverted)
        
        # again normal
        one.process()
        self.assertEqual(one.to_dict(), expect_processed)
        self.assertEqual(one.to_dict(revert=True), expect_reverted)
        
        # from bulk item -------------------------------------------------------
        bulk_one = ItemGeneralOne.Bulk(f_decimal='1000!1000')
        # bulk
        for i, j in zip(range(1, 3), range(3, 5)):
            bulk_one.gen(f_integer=i*10,
                         f_decimal='{}!{}'.format(i*10, i*10),
                         f_date='20.10.2010',
                         f_time='10-20-30',
                         f_datetime='20.10.2010 10-20-30')
            bulk_one['two_x_x'].gen(f_integer=j*10,
                                    f_decimal='{}.{}'.format(j*10, j*10),
                                    f_date='2010-10-20',
                                    f_time='10:20:30',
                                    f_datetime='2010-10-20 10:20:30')

        expect_processed = {
             'defaults': {
                 'two_x_x': {
                     'defaults': {},
                     'bulk': [{
                         'item': {
                             'f_integer': 30,
                             'f_date': datetime.date(2010, 10, 20),
                             'f_time': datetime.time(10, 20, 30),
                             'one_x_x': {
                                 'defaults': {
                                     'two_x_x': {
                                         'defaults': {},
                                         'bulk': [{
                                             'id': 1
                                         }]
                                     }
                                 },
                                 'bulk': [{
                                     'item': {
                                         'f_integer': 10,
                                         'f_date': datetime.date(2010, 10, 20),
                                         'f_time': datetime.time(10, 20, 30),
                                         'f_decimal': Decimal('10.10'),
                                         'two_x_x': {
                                             'defaults': {},
                                             'bulk': [{
                                                 'id': 1
                                             }, {
                                                 'item': {
                                                     'f_integer': 40,
                                                     'f_date':datetime.date(2010, 10, 20),
                                                     'f_time': datetime.time(10, 20, 30),
                                                     'one_x_x': {
                                                         'defaults': {
                                                             'two_x_x': {
                                                                 'defaults': {},
                                                                 'bulk': [{
                                                                     'id': 2
                                                                 }]
                                                             }
                                                         },
                                                         'bulk': [{
                                                             'id': 3
                                                         }, {
                                                             'item': {
                                                                 'f_integer': 20,
                                                                 'f_date': datetime.date(2010, 10, 20),
                                                                 'f_time': datetime.time(10, 20, 30),
                                                                 'f_decimal': Decimal('20.20'),
                                                                 'two_x_x': {
                                                                     'defaults': {},
                                                                     'bulk': [{
                                                                         'id': 1
                                                                     }, {
                                                                         'id': 2
                                                                     }]
                                                                 },
                                                                 'f_datetime': datetime.datetime(2010, 10, 20, 10, 20, 30)
                                                             },
                                                             'id': 4
                                                         }]
                                                     },
                                                     'f_decimal': Decimal('40.40'),
                                                     'f_datetime': datetime.datetime(2010, 10, 20, 10, 20, 30)
                                                 },
                                                 'id': 2
                                             }]
                                         },
                                         'f_datetime': datetime.datetime(2010, 10, 20, 10, 20, 30)
                                     },
                                     'id': 3
                                 }, {
                                     'id': 4
                                 }]
                             },
                             'f_decimal': Decimal('30.30'),
                             'f_datetime': datetime.datetime(2010, 10, 20, 10, 20, 30)
                         },
                         'id': 1
                     }, {
                         'id': 2
                     }]
                 },
                 'f_decimal': Decimal('1000.1000')
             },
             'bulk': [{
                 'id': 3
             }, {
                 'id': 4
             }]
         }
        
        bulk_one.process()
        self.assertEqual(bulk_one.to_dict(), expect_processed)
        
        expect_reverted = {
            'defaults': {
                'two_x_x': {
                    'defaults': {},
                    'bulk': [{
                        'item': {
                            'one_x_x': {
                                'defaults': {
                                    'two_x_x': {
                                        'defaults': {},
                                        'bulk': [{
                                            'id': 1
                                        }]
                                    }
                                },
                                'bulk': [{
                                    'item': {
                                        'f_integer': 10,
                                        'f_date': '20.10.2010',
                                        'f_decimal': '10!10',
                                        'f_time': '10-20-30',
                                        'two_x_x': {
                                            'defaults': {},
                                            'bulk': [{
                                                'id': 1
                                            }, {
                                                'item': {
                                                    'one_x_x': {
                                                        'defaults': {
                                                            'two_x_x': {
                                                                'defaults': {},
                                                                'bulk': [{
                                                                    'id': 2
                                                                }]
                                                            }
                                                        },
                                                        'bulk': [{
                                                            'id': 3
                                                        }, {
                                                            'item': {
                                                                'f_integer': 20,
                                                                'f_date': '20.10.2010',
                                                                'f_decimal': '20!20',
                                                                'f_time': '10-20-30',
                                                                'two_x_x': {
                                                                    'defaults': {},
                                                                    'bulk': [{
                                                                        'id': 1
                                                                    }, {
                                                                        'id': 2
                                                                    }]
                                                                },
                                                                'f_datetime': '20.10.2010 10-20-30'
                                                            },
                                                            'id': 4
                                                        }]
                                                    },
                                                    'f_integer': 40,
                                                    'f_date': '2010-10-20',
                                                    'f_decimal': '40.40',
                                                    'f_time': '10:20:30',
                                                    'f_datetime': '2010-10-20 10:20:30'
                                                },
                                                'id': 2
                                            }]
                                        },
                                        'f_datetime': '20.10.2010 10-20-30'
                                    },
                                    'id': 3
                                }, {
                                    'id': 4
                                }]
                            },
                            'f_integer': 30,
                            'f_date': '2010-10-20',
                            'f_decimal': '30.30',
                            'f_time': '10:20:30',
                            'f_datetime': '2010-10-20 10:20:30'
                        },
                        'id': 1
                    }, {
                        'id': 2
                    }]
                },
                'f_decimal': '1000!1000'
            },
            'bulk': [{
                'id': 3
            }, {
                'id': 4
            }]
        }
        
        self.assertEqual(bulk_one.to_dict(revert=True), expect_reverted)
        # original not changed
        self.assertEqual(bulk_one.to_dict(), expect_processed)
        
        # all reverted
        bulk_one.revert()
        self.assertEqual(bulk_one.to_dict(revert=True), expect_reverted)
        self.assertEqual(bulk_one.to_dict(), expect_reverted)
        
        # back again
        bulk_one.process()
        self.assertEqual(bulk_one.to_dict(), expect_processed)
        self.assertEqual(bulk_one.to_dict(revert=True), expect_reverted)
    
    
    def test_multiple_date_time_formats_to_dict_revert(self):
        self.item_cls_manager.clear()
        self.item_cls_manager.autogenerate = True
        
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            conversions = {
                'date_formats': ['%d.%m.%Y', '%d.%m.%y'],
                'time_formats': ['%H:%M:%S', '%H:%M'],
                'datetime_formats': ['%d.%m.%Y %H:%M:%S', '%d.%m.%y %H:%M'],
            }
        
        # long format versions
        first_item = ItemGeneralOne(f_date='20.10.2010',
                                    f_time='10:20:30',
                                    f_datetime='20.10.2010 10:20:30')
        first_item.process()
        expect = {
            'item': {
                'f_time': datetime.time(10, 20, 30),
                'f_date': datetime.date(2010, 10, 20),
                'f_datetime': datetime.datetime(2010, 10, 20, 10, 20, 30),
            }
        }
        self.assertEqual(first_item.to_dict(), expect)
        
        first_item.revert()
        expect = {
            'item': {
                'f_time': '10:20:30',
                'f_datetime': '20.10.2010 10:20:30',
                'f_date': '20.10.2010',
            }
        }
        self.assertEqual(first_item.to_dict(), expect)
        
        # short format versions
        first_item = ItemGeneralOne(f_date='20.10.10',
                                    f_time='10:20',
                                    f_datetime='20.10.10 10:20')
        first_item.process()
        expect = {
            'item': {
                'f_time': datetime.time(10, 20, 0),
                'f_date': datetime.date(2010, 10, 20),
                'f_datetime': datetime.datetime(2010, 10, 20, 10, 20, 0),
            }
        }
        self.assertEqual(first_item.to_dict(), expect)
        
        first_item.revert()  # long versions will be used
        expect = {
            'item': {
                'f_time': '10:20:00',
                'f_datetime': '20.10.2010 10:20:00',
                'f_date': '20.10.2010',
            }
        }
        self.assertEqual(first_item.to_dict(), expect)

            


from save_to_db.core.item import Item
from save_to_db.utils.test_base import TestBase


class TestDictWrapper(TestBase):
    
    def test_wrap_and_get_item(self):
        
        class TestItem(Item):
            pass
        
        original_item = TestItem()
        dict_wrapper = original_item.dict_wrapper()
        returned_item = dict_wrapper.get_item()
        
        self.assertIs(returned_item, returned_item)


from save_to_db.adapters.utils.adapter_manager import get_adapter_cls
from save_to_db.adapters.utils.column_type import ColumnType
from save_to_db.adapters.utils.relation_type import RelationType
from save_to_db.exceptions import (
    InvalidFieldName,
    RelatedItemNotFound,
    MultipleRelatedItemsFound,
    NonExistentRelationDefined,
    NonExistentFieldUsed,
    WrongAlias,
    FieldsFromRelationNotAllowed,
    NorewriteRelationException,
    NorewriteKeyUsedTwice,
    DeleterSelectorsOrKeepersEmpty,
    DeleterXToManyRelationUsed,
    UnrefNonXToManyFieldUsed)
from save_to_db.core.item import Item
from save_to_db.utils.test_base import TestBase


class TestItemSetup(TestBase):
    
    ModelFieldTypes = None
    
    ModelGeneralOne = None
    ModelGeneralTwo = None
    
    ModelInvalidFieldNames = None
    
    ModelAutoReverseOne = None
    ModelAutoReverseTwoA = None
    ModelAutoReverseTwoB = None
    ModelAutoReverseThreeA = None
    ModelAutoReverseThreeB = None
    ModelAutoReverseFourA = None
    ModelAutoReverseFourB = None
    
    ModelConstraintsOne = None
    ModelConstraintsTwo = None
    ModelConstraintsThree = None
    ModelConstraintsFour = None
    ModelConstraintsFive = None
    ModelConstraintsSix = None
    ModelConstraintsSelf = None
    
    
    def setup_models(self, one=True, two=True, one_conf=None, two_conf=None,
                     clear_registry=False, complete_setup=None):
        if clear_registry:
            self.item_cls_manager.clear()
        
        if one:
            dct = {'model_cls': self.ModelGeneralOne, }
            if one_conf:
                dct.update(one_conf)
            self.ItemGeneralOne = type('ItemGeneralOne', (Item,), dct)
        if two:
            dct = {'model_cls': self.ModelGeneralTwo, }
            if two_conf:
                dct.update(two_conf)
            self.ItemGeneralTwo = type('ItemGeneralTwo', (Item,), dct)
        
        if complete_setup or (complete_setup is None and one and two):
            self.ItemGeneralOne.complete_setup()
            self.ItemGeneralTwo.complete_setup()
    
    
    def test_auto_reverse_relations(self):
        class ItemAutoReverseOne(Item):
            model_cls = self.ModelAutoReverseOne
        
        class ItemAutoReverseTwoA(Item):
            model_cls = self.ModelAutoReverseTwoA
        
        class ItemAutoReverseTwoB(Item):
            model_cls = self.ModelAutoReverseTwoB
        
        class ItemAutoReverseThreeA(Item):
            model_cls = self.ModelAutoReverseThreeA
        
        class ItemAutoReverseThreeB(Item):
            model_cls = self.ModelAutoReverseThreeB
        
        class ItemAutoReverseFourA(Item):
            model_cls = self.ModelAutoReverseFourA
        
        class ItemAutoReverseFourB(Item):
            model_cls = self.ModelAutoReverseFourB


        for item_cls in (ItemAutoReverseOne,
                         ItemAutoReverseTwoA, ItemAutoReverseTwoB,
                         ItemAutoReverseThreeA, ItemAutoReverseThreeB,
                         ItemAutoReverseFourA, ItemAutoReverseFourB,):
            item_cls()  # completes relations
            relations_and_reverse = {}
            
            for key, relations in item_cls.relations.items():
                relations_and_reverse[key] = (
                    relations['reverse_key'],
                    relations['relation_type'],
                )
            
            self.assertEqual(relations_and_reverse,
                             item_cls.model_cls.ITEM_RELATIONS)
    
    
    def test_nonexistent_field(self):
        # proper configuration (no nonexistent fields)
        self.setup_models(
            one_conf={
                'creators': ['f_text', 'child_1_1'],  # field and relation
                'getters': ['f_float', 'parent_1_1'],
                'nullables': ['f_date', 'child_1_x'],
                'defaults': {'f_integer': None},
                'norewrite_fields': {'f_integer': True},
                'unref_x_to_many': {
                    'two_x_x': {
                        'selectors': ['f_integer', 'f_float'],
                        'keepers': ['f_float', 'one_1_1']
                    },
                    'two_1_x': ['f_integer', 'f_text'],
                    'child_1_x': {
                        'selectors': ['f_integer', 'f_float', 'two_1_1'],
                    },
                },
                'remove_null_fields': ['f_text', 'child_1_1'],
            },
            clear_registry=True)
        item_one_cls = self.ItemGeneralOne
        self.assertEqual(item_one_cls.creators, [{'f_text'}, {'child_1_1'}])
        self.assertEqual(item_one_cls.getters, [{'f_float'}, {'parent_1_1'}])
        self.assertEqual(item_one_cls.nullables, {'child_1_x', 'f_date'})
        self.assertEqual(item_one_cls.remove_null_fields,
                         {'f_text', 'child_1_1'})
        self.assertEqual(item_one_cls.norewrite_fields, {'f_integer': True})
        
        # nonexistent field
        for key in ('creators', 'getters', 'nullables', 'remove_null_fields',
                    'deleter_selectors', 'deleter_keepers'):
            with self.assertRaises(NonExistentFieldUsed):
                self.setup_models(two_conf={key: ['nonexistent']},
                                  clear_registry=True)
        
        with self.assertRaises(NonExistentFieldUsed):
            self.setup_models(two_conf={'defaults': {'nonexistent': None}},
                              clear_registry=True)
        
        with self.assertRaises(NonExistentFieldUsed):
            self.setup_models(
                two_conf={
                    'norewrite_fields': {
                        'nonexistent': True
                    }
                },
                clear_registry=True)
        
        with self.assertRaises(NonExistentFieldUsed):
            self.setup_models(
                two_conf={
                    'unref_x_to_many': {
                        'nonexistent': ['field_1', 'field_2']
                    }
                },
                clear_registry=True)
        
        with self.assertRaises(NonExistentFieldUsed):
            self.setup_models(
                two_conf={
                    'unref_x_to_many': {
                        'one_x_x': ['nonexistent']
                    }
                },
                clear_registry=True)
           
        with self.assertRaises(NonExistentFieldUsed):
            self.setup_models(
                two_conf={
                    'unref_x_to_many': {
                        'one_x_x': {
                            'keepers': ['nonexistent'],
                        }
                    }
                },
                clear_registry=True)
    
    
    def test_unref_x_to_many_non_x_to_many_field(self):
        # proper configuration
        self.setup_models(
            one_conf={
                'unref_x_to_many': {
                    'two_x_x': {
                        'selectors': ['one_x_1'],
                        'keepers': ['one_1_1']
                    },
                    'two_1_x': ['one_x_1', 'one_1_1'],
                    'child_1_x': {
                        'selectors': ['parent_x_1', 'parent_1_1'],
                    },
                }
            },
            clear_registry=True)
        
        with self.assertRaises(UnrefNonXToManyFieldUsed):
            self.setup_models(
                two_conf={
                    'unref_x_to_many': {
                        'f_integer': ['field']
                    }
                },
                clear_registry=True)
        
        with self.assertRaises(UnrefNonXToManyFieldUsed):
            self.setup_models(
                two_conf={
                    'unref_x_to_many': {
                        'one_x_1': ['f_integer']
                    }
                },
                clear_registry=True)
        
        with self.assertRaises(UnrefNonXToManyFieldUsed):
            self.setup_models(
                two_conf={
                    'unref_x_to_many': {
                        'one_1_1': ['f_integer']
                    }
                },
                clear_registry=True)
    
    
    def test_norewrite_relations(self):
        self.setup_models(
            one_conf={
                'norewrite_fields': {
                    'f_integer': True,
                    'two_1_1': True,
                    'two_x_1': True,
                    'two_x_x': True,
                },
            },
            two_conf={
                'norewrite_fields': {
                    'f_integer': False,
                    'one_x_1': False,
                },
            },
            clear_registry=True)
        
        expected = {
            'f_integer': True,
            'two_1_1': True,
            'two_x_1': True,
            'two_x_x': True,
            'two_1_x': False,  # reverse from two
        }
        self.assertEqual(self.ItemGeneralOne.norewrite_fields, expected)
        
        expected = {
            'f_integer': False,
            'one_1_1': True, # reverse from one
            'one_x_1': False,
            'one_1_x': True,  # reverse from one
            'one_x_x': True,  # reverse from one
        }
        self.assertEqual(self.ItemGeneralTwo.norewrite_fields, expected)
        
        with self.assertRaises(NorewriteRelationException):
            self.setup_models(
                one_conf={
                    'norewrite_fields': {
                        'two_1_1': True,
                    },
                },
                two_conf={
                    'norewrite_fields': {
                        'one_1_1': False,
                    },
                },
                clear_registry=True)
    
    
    def test_norewrite_for_general_key(self):
        class ItemGeneralOne(Item):
            model_cls = self.ModelGeneralOne
            norewrite_fields = {
                'f_integer': True,
                True: False,
            }
         
        class ItemGeneralTwo(Item):
            model_cls = self.ModelGeneralTwo
        
        # first checking item two, it must have values from item one
        ItemGeneralTwo()
        expect = {
            'one_x_x': False,
            'one_1_x': False,
            'one_1_1': False,
            'one_x_1': False,
        }
        self.assertEqual(ItemGeneralTwo.norewrite_fields, expect)
        
        ItemGeneralOne()
        expect = {
            'f_integer': True,
            'f_binary': False,
            'f_date': False,
            'parent_x_x': False,
            'two_1_1': False,
            'f_text': False,
            'child_1_1': False,
            'two_x_1': False,
            'id': False,
            'child_x_x': False,
            'two_1_x': False,
            'f_boolean': False,
            'f_string': False,
            'parent_1_1': False,
            'f_datetime': False,
            'child_1_x': False,
            'f_float': False,
            'f_decimal': False,
            'two_x_x': False,
            'f_time': False,
            'parent_x_1': False}
        self.assertEqual(ItemGeneralOne.norewrite_fields, expect)
    
    
    def test_field_from_relations_not_allowed(self):
        # proper configuration (no nonexistent fields)
        self.setup_models(
            one_conf={
                'creators': ['f_text', 'child_1_1'],  # field and relation
                'getters': ['f_float', 'parent_1_1'],
                'nullables': ['f_date', 'child_1_x'],
                'defaults': {'f_integer': None},
                'norewrite_fields': {'f_integer': True},
                'remove_null_fields': ['f_text', 'child_1_1'],
            },
            two_conf={
                'creators': ['f_integer'],
                'getters': ['child_1_1'],
            },           
            clear_registry=True)
        item_one_cls = self.ItemGeneralOne
        item_two_cls = self.ItemGeneralTwo
        
        self.assertEqual(item_one_cls.creators, [{'f_text'}, {'child_1_1'}])
        self.assertEqual(item_one_cls.getters, [{'f_float'}, {'parent_1_1'}])
        self.assertEqual(item_one_cls.nullables, {'child_1_x', 'f_date'})
        self.assertEqual(item_one_cls.norewrite_fields, {'f_integer': True})
        self.assertEqual(item_one_cls.remove_null_fields,
                         {'f_text', 'child_1_1'})
        self.assertEqual(item_two_cls.creators, [{'f_integer'}])
        self.assertEqual(item_two_cls.getters, [{'child_1_1'}])
        
         
        # field_from_relations_not_allowed
        for key in ('creators', 'getters',):
            with self.assertRaises(FieldsFromRelationNotAllowed):
                self.setup_models(two_conf={key: [{'child_1_1__f_integer'}]},
                                  clear_registry=True)
        
        for key in ('nullables', 'remove_null_fields',
                    'deleter_selectors', 'deleter_keepers'):
            with self.assertRaises(FieldsFromRelationNotAllowed):
                self.setup_models(two_conf={key: ['child_1_1__f_integer']},
                                  clear_registry=True)
        
        with self.assertRaises(FieldsFromRelationNotAllowed):
            self.setup_models(
                two_conf={
                    'defaults': {
                        'child_1_1__f_integer': None
                    }
                },
                clear_registry=True)
            
        with self.assertRaises(FieldsFromRelationNotAllowed):
            self.setup_models(
                two_conf={
                    'norewrite_fields': {
                        'child_1_1__f_integer': True
                    }
                },
                clear_registry=True)
        
        with self.assertRaises(FieldsFromRelationNotAllowed):
            self.setup_models(
                one_conf={
                    'unref_x_to_many': {
                        'two_x_x__one_x_x': ['f_integer']
                    }
                },
                clear_registry=True)
        
        with self.assertRaises(FieldsFromRelationNotAllowed):
            self.setup_models(
                one_conf={
                    'unref_x_to_many': {
                        'two_x_x': ['one_x_x__f_integer']
                    }
                },
                clear_registry=True)
    
    
    def test_norewrite_fields_config(self):
        self.setup_models(
            one_conf={
                'norewrite_fields': {
                    'f_integer': True,
                    True: False,
                },
            },           
            clear_registry=True)
        
        expect = {
            'id': False,
            'f_binary': False,
            'f_boolean': False,
            'f_date': False,
            'f_datetime': False,
            'f_float': False,
            'f_decimal': False,
            'f_integer': True, 
            'f_string': False,
            'f_text': False, 
            'f_time': False,
            'child_1_1': False,
            'child_1_x': False,
            'child_x_x': False,
            'parent_1_1': False,
            'parent_x_1': False,
            'parent_x_x': False,
            'two_1_1': False,
            'two_1_x': False,
            'two_x_1': False,
            'two_x_x': False,
        }
        
        self.assertEqual(self.ItemGeneralOne.norewrite_fields, expect)
        
        self.setup_models(
            one_conf={
                'norewrite_fields': {
                    'f_integer': True,
                    # tuple
                    ('f_string', 'f_text'): None,
                    # RelationType
                    RelationType.MANY_TO_MANY: True,
                    RelationType.ONE_TO_ONE: True,
                    # RelationType overwritten
                    'parent_x_x': None,
                    # the rest
                    True: False,
                },
            },           
            clear_registry=True)
        
        excpext = {
            'f_integer': True,      # 'f_integer
            'f_binary': False,      # True
            'parent_x_1': False,    # True
            'f_datetime': False,    # True
            'f_text': None,         # ('f_string', 'f_text')
            'f_string': None,       # ('f_string', 'f_text')
            'two_1_x': False,       # True
            'two_x_x': True,        # RelationType.MANY_TO_MANY
            'child_x_x': None,      # 'parent_x_x' (reverse)
            'parent_1_1': True,     # RelationType.ONE_TO_ONE
            'id': False,            # True
            'f_boolean': False,     # True
            'f_float': False,       # True
            'f_decimal': False,     # True
            'two_x_1': False,       # True
            'f_time': False,        # True
            'child_1_1': True,      # RelationType.ONE_TO_ONE
            'two_1_1': True,        # RelationType.ONE_TO_ONE
            'child_1_x': False,     # True
            'f_date': False,        # True
            'parent_x_x': None,     # 'parent_x_x'
        }
        self.assertEqual(self.ItemGeneralOne.norewrite_fields, excpext)
    
    
    def test_no_rewrite_field_used_twice(self):
        with self.assertRaises(NorewriteKeyUsedTwice):
            self.setup_models(
            one_conf={
                'norewrite_fields': {
                    'f_integer': True,
                    ('f_integer', 'f_float'): None,
                },
            },           
            clear_registry=True)
        
        with self.assertRaises(NorewriteKeyUsedTwice):
            self.setup_models(
            one_conf={
                'norewrite_fields': {
                    ('f_string', 'f_integer'): True,
                    ('f_integer', 'f_float'): None,
                },
            },           
            clear_registry=True)
    
    
    def test_relations_config_excpetions(self):
        # NonExistentRelationDefined
        class WithNonExistantRelation(Item):
            model_cls = self.ModelFieldTypes
            relations = {
                'non_existant': self.ModelGeneralOne,
            }

        with self.assertRaises(NonExistentRelationDefined):
            WithNonExistantRelation.complete_setup()
        
        # RelatedItemNotFound
        self.setup_models(one=True, two=False)
        with self.assertRaises(RelatedItemNotFound):
            self.ItemGeneralOne.complete_setup()
        
        # MultipleRelatedItemsFound
        self.setup_models(one=False, two=True)
        self.setup_models(one=False, two=True)  # second model
        
        with self.assertRaises(MultipleRelatedItemsFound):
            self.ItemGeneralOne.complete_setup()
    

    def test_fields_autoconfig(self):
        self.setup_models()
        
        extected_fields = {
            'id': ColumnType.INTEGER,
            'f_binary': ColumnType.BINARY,
            'f_boolean': ColumnType.BOOLEAN,
            'f_text': ColumnType.TEXT,
            'f_string': ColumnType.STRING,
            'f_integer': ColumnType.INTEGER,
            'f_float': ColumnType.FLOAT,
            'f_decimal': ColumnType.DECIMAL,
            'f_date': ColumnType.DATE,
            'f_time': ColumnType.TIME,
            'f_datetime': ColumnType.DATETIME,
        }
        self.assertEqual(self.ItemGeneralOne.fields, extected_fields)
    
    
    def test_relations_autoconfig(self):
        self.setup_models(two_conf={
            'relations': {
                    'parent_x_x': {
                        'replace_x_to_many': True,
                    },
                },
            })
        
        #--- one ---
        expected_relations = {
            'parent_1_1': {
                'relation_type': RelationType.ONE_TO_ONE,
                'item_cls': self.ItemGeneralOne,
                'replace_x_to_many': False,
                'reverse_key': 'child_1_1',
            },
            'child_1_1': {
                'relation_type': RelationType.ONE_TO_ONE,
                'item_cls': self.ItemGeneralOne,
                'replace_x_to_many': False,
                'reverse_key': 'parent_1_1',
            },
            'parent_x_1': {
                'relation_type': RelationType.MANY_TO_ONE,
                'item_cls': self.ItemGeneralOne,
                'replace_x_to_many': False,
                'reverse_key': 'child_1_x',
            },
            'child_1_x': {
                'relation_type': RelationType.ONE_TO_MANY,
                'item_cls': self.ItemGeneralOne,
                'replace_x_to_many': False,
                'reverse_key': 'parent_x_1',
            },
            'parent_x_x': {
                'relation_type': RelationType.MANY_TO_MANY,
                'item_cls': self.ItemGeneralOne,
                'replace_x_to_many': False,
                'reverse_key': 'child_x_x',
            },
            'child_x_x': {
                'relation_type': RelationType.MANY_TO_MANY,
                'item_cls': self.ItemGeneralOne,
                'replace_x_to_many': False,
                'reverse_key': 'parent_x_x',
            },
            'two_1_1': {
                'relation_type': RelationType.ONE_TO_ONE,
                'item_cls': self.ItemGeneralTwo,
                'replace_x_to_many': False,
                'reverse_key': 'one_1_1',
            },
            'two_1_x': {
                'relation_type': RelationType.ONE_TO_MANY,
                'item_cls': self.ItemGeneralTwo,
                'replace_x_to_many': False,
                'reverse_key': 'one_x_1',
            },
            'two_x_1': {
                'relation_type': RelationType.MANY_TO_ONE,
                'item_cls': self.ItemGeneralTwo,
                'replace_x_to_many': False,
                'reverse_key': 'one_1_x',
            },
            'two_x_x': {
                'relation_type': RelationType.MANY_TO_MANY,
                'item_cls': self.ItemGeneralTwo,
                'replace_x_to_many': False,
                'reverse_key': 'one_x_x',
            },
        }
        self.assertEqual(self.ItemGeneralOne.relations, expected_relations)
        
        #--- two ---
        expected_relations = {
            'parent_1_1': {
                'relation_type': RelationType.ONE_TO_ONE,
                'item_cls': self.ItemGeneralTwo,
                'replace_x_to_many': False,
                'reverse_key': 'child_1_1',
            },
            'child_1_1': {
                'relation_type': RelationType.ONE_TO_ONE,
                'item_cls': self.ItemGeneralTwo,
                'replace_x_to_many': False,
                'reverse_key': 'parent_1_1',
            },
            'parent_x_1': {
                'relation_type': RelationType.MANY_TO_ONE,
                'item_cls': self.ItemGeneralTwo,
                'replace_x_to_many': False,
                'reverse_key': 'child_1_x',
            },
            'child_1_x': {
                'relation_type': RelationType.ONE_TO_MANY,
                'item_cls': self.ItemGeneralTwo,
                'replace_x_to_many': False,
                'reverse_key': 'parent_x_1',
            },
            'parent_x_x': {
                'relation_type': RelationType.MANY_TO_MANY,
                'item_cls': self.ItemGeneralTwo,
                'replace_x_to_many': True,  # default overwritten
                'reverse_key': 'child_x_x',
            },
            'child_x_x': {
                'relation_type': RelationType.MANY_TO_MANY,
                'item_cls': self.ItemGeneralTwo,
                'replace_x_to_many': False,
                'reverse_key': 'parent_x_x',
            },
            'one_1_1': {
                'relation_type': RelationType.ONE_TO_ONE,
                'item_cls': self.ItemGeneralOne,
                'replace_x_to_many': False,
                'reverse_key': 'two_1_1',
            },
            'one_1_x': {
                'relation_type': RelationType.ONE_TO_MANY,
                'item_cls': self.ItemGeneralOne,
                'replace_x_to_many': False,
                'reverse_key': 'two_x_1',
            },
            'one_x_1': {
                'relation_type': RelationType.MANY_TO_ONE,
                'item_cls': self.ItemGeneralOne,
                'replace_x_to_many': False,
                'reverse_key': 'two_1_x',
            },
            'one_x_x': {
                'relation_type': RelationType.MANY_TO_MANY,
                'item_cls': self.ItemGeneralOne,
                'replace_x_to_many': False,
                'reverse_key': 'two_x_x',
            },
        }
        self.assertEqual(self.ItemGeneralTwo.relations, expected_relations)
    
    
    def __to_sorted_lists(self, iterable_of_iterables):
        result = []
        for iterable in iterable_of_iterables:
            element = list(iterable)
            element.sort()
            result.append(element)
        result.sort()
        return result
    
    
    def test_creators_getter_autoconfig(self):
        # generating items for models
        models_for_items = (
            self.ModelConstraintsOne,
            self.ModelConstraintsTwo,
            self.ModelConstraintsThree,
            self.ModelConstraintsFour,
            self.ModelConstraintsFive,
            self.ModelConstraintsSix,
            self.ModelConstraintsSelf,
        )
        item_classes = []
        for model_cls in models_for_items:
            item_classes.append(type(
                model_cls.__name__.replace('Model', 'Item'), (Item,), {
                    'model_cls': model_cls
                }
            ))
            
        for item_cls in item_classes:
            item_cls.complete_setup()
        (ItemConstraintsOne, ItemConstraintsTwo, ItemConstraintsThree,
         ItemConstraintsFour, ItemConstraintsFive, ItemConstraintsSix,
         ItemConstraintsSelf) = item_classes
        
        # checking configuration
        adapter_cls = get_adapter_cls(ItemConstraintsOne.model_cls)
        composites = adapter_cls.COMPOSITE_KEYS_SUPPORTED
        
        expected = {
            ItemConstraintsOne: {
                'creators': [{'f_text', 'f_string'}],
                'getters': [{'id'}, {'f_integer'}, {'f_string'}, {'five_1_1'}],
            },
            ItemConstraintsTwo: {
                'creators': [{'four_x_1'}],
                'getters': [{'id'}],
            },
            ItemConstraintsThree: {
                'creators': [{'one_x_1'}],
                'getters': [{'id'}],
            },
            ItemConstraintsFour: {
                'creators':
                    [{'primary_two', 'primary_one'}]
                        if composites else [],
                'getters':
                    [{'primary_two', 'primary_one'} if composites else {'id'},
                     {'five_1_1'}, {'f_integer', 'f_string'}],
            },
            ItemConstraintsFive: {
                'creators': [],
                'getters': [{'id'}, {'one_1_1'}, {'four_1_1'}, {'six_1_1'}],
            },
            ItemConstraintsSix: {
                'creators': [{'five_1_1'}],
                'getters': [{'id'}, {'f_integer', 'five_1_1'}, {'five_1_1'}],
            },
            ItemConstraintsSelf: {
                'creators': [{'code', 'second_parent_1_1', 'parent_x_1'}],
                'getters': [{'code'},
                            {'first_parent_1_1'}, {'first_child_1_1'},
                            {'second_parent_1_1'}, {'second_child_1_1'},],
            },
        }
        
        for item_cls in item_classes:
            #--- getters ---
            item_getters = self.__to_sorted_lists(
                item_cls.getters)
            expected_getters = self.__to_sorted_lists(
                expected[item_cls]['getters'])
            self.assertEqual(item_getters, expected_getters, item_cls)
            
            #--- creators ---
            item_creators = self.__to_sorted_lists(
                item_cls.creators)
            expected_creators = self.__to_sorted_lists(
                expected[item_cls]['creators'])
            
            self.assertEqual(item_creators, expected_creators, item_cls)
    
    
    def test_creators_getter_autoconfig_overwrite(self):
        # default overwritten
        class ItemConstraintsSelf_A(Item):
            model_cls = self.ModelConstraintsSelf
            creators = [{'parent_x_1'}]
            getters = [{'first_parent_1_1'}]
        ItemConstraintsSelf_A.complete_setup()
        
        self.assertEqual(ItemConstraintsSelf_A.creators,
                         [{'parent_x_1'}])
        self.assertEqual(ItemConstraintsSelf_A.getters,
                         [{'first_parent_1_1'}])
    

        # default not overwritten
        self.item_cls_manager.clear()
        class ItemConstraintsSelf_B(Item):
            model_cls = self.ModelConstraintsSelf
        ItemConstraintsSelf_B.complete_setup()
        
        self.assertEqual(self.__to_sorted_lists(ItemConstraintsSelf_B.creators),
                         self.__to_sorted_lists(
                             [{'code', 'second_parent_1_1', 'parent_x_1'}]))
        self.assertEqual(self.__to_sorted_lists(ItemConstraintsSelf_B.getters),
                         self.__to_sorted_lists(
                            [{'code'},
                             {'first_parent_1_1'}, {'first_child_1_1'},
                             {'second_parent_1_1'}, {'second_child_1_1'},]))
        
        # default creators merge
        self.item_cls_manager.clear()
        class ItemConstraintsSelf_C(Item):
            model_cls = self.ModelConstraintsSelf
            creators = [{'f_binary'}]
            creators_autoconfig = True
            autoinject_creators = False
            getters = [{'f_boolean'}]
            getters_autoconfig = False
        ItemConstraintsSelf_C.complete_setup()
        
        self.assertEqual(self.__to_sorted_lists(ItemConstraintsSelf_C.creators),
                         self.__to_sorted_lists(
                            [{'f_binary'},
                             {'code', 'second_parent_1_1', 'parent_x_1'}]))
        self.assertEqual(self.__to_sorted_lists(ItemConstraintsSelf_C.getters),
                         self.__to_sorted_lists([{'f_boolean'}]))
        
        # default getters merge
        self.item_cls_manager.clear()
        class ItemConstraintsSelf_D(Item):
            model_cls = self.ModelConstraintsSelf
            creators = [{'f_binary'}]
            creators_autoconfig = False
            autoinject_creators = False
            getters = [{'f_boolean'}]
            getters_autoconfig = True
        ItemConstraintsSelf_D.complete_setup()
        
        self.assertEqual(self.__to_sorted_lists(ItemConstraintsSelf_D.creators),
                         self.__to_sorted_lists([{'f_binary'}]))
        self.assertEqual(self.__to_sorted_lists(ItemConstraintsSelf_D.getters),
                         self.__to_sorted_lists(
                            [{'f_boolean'},
                             {'code'},
                             {'first_parent_1_1'}, {'first_child_1_1'},
                             {'second_parent_1_1'}, {'second_child_1_1'},]))
        
        # creators extended
        self.item_cls_manager.clear()
        class ItemConstraintsSelf_E(Item):
            model_cls = self.ModelConstraintsSelf
        ItemConstraintsSelf_E.complete_setup()
        
        self.assertEqual(self.__to_sorted_lists(ItemConstraintsSelf_E.creators),
                         self.__to_sorted_lists([{'code',
                                                  'second_parent_1_1',
                                                  'parent_x_1'}]))
        
        self.item_cls_manager.clear()
        class ItemConstraintsSelf_F(Item):
            model_cls = self.ModelConstraintsSelf
            creators = [{'f_binary'}, {'f_boolean', 'code'}]
            creators_autoconfig = False
            autoinject_creators = True  # prevented by `autoinject_creators`
        ItemConstraintsSelf_F.complete_setup()
        
        self.assertEqual(self.__to_sorted_lists(ItemConstraintsSelf_F.creators),
                         self.__to_sorted_lists([{'f_binary'},
                                                 {'f_boolean', 'code'}]))
        
        self.item_cls_manager.clear()
        class ItemConstraintsSelf_G(Item):
            model_cls = self.ModelConstraintsSelf
            creators = [{'f_binary'}, {'f_boolean', 'code'}]
            creators_autoconfig = True
            autoinject_creators = True  # creator groups must be extended
        ItemConstraintsSelf_G.complete_setup()
         
        self.assertEqual(self.__to_sorted_lists(ItemConstraintsSelf_G.creators),
                         self.__to_sorted_lists([{'f_binary',
                                                  'code',
                                                  'second_parent_1_1',
                                                  'parent_x_1'},
                                                 {'f_boolean',
                                                  'code',
                                                  'second_parent_1_1',
                                                  'parent_x_1'}]))
        
        # but do not extend if group already there
        self.item_cls_manager.clear()
        class ItemConstraintsSelf_H(Item):
            model_cls = self.ModelConstraintsSelf
            creators = [{'code',
                         'second_parent_1_1', 'parent_x_1'}]
            creators_autoconfig = True
            autoinject_creators = True
        ItemConstraintsSelf_H.complete_setup()
         
        self.assertEqual(self.__to_sorted_lists(ItemConstraintsSelf_H.creators),
                         self.__to_sorted_lists([{'code',
                                                  'second_parent_1_1',
                                                  'parent_x_1'}]))
    
    
    def test_invalid_field_name(self):
        class InvalidItem(Item):
            model_cls = self.ModelInvalidFieldNames

        with self.assertRaises(InvalidFieldName):
            InvalidItem()
    
    
    def test_invalid_alias(self):
        with self.assertRaises(WrongAlias):
            self.setup_models(
                one_conf={
                    'aliases': {
                        'alias_1': 'two_1_1__f_integer',
                        'alias_2': 'two_x_x__one_1_1__f_integer__invaliad_alias',
                    }
                },
                clear_registry=True)
        
        
    def test_creators_and_getters_mix(self):
        self.item_cls_manager.autogenerate = True
        
        class ItemConstraintsOne(Item):
            model_cls = self.ModelConstraintsOne
            
            getters = ['three_1_x', 'f_integer']
            getters_autoconfig = True
            
            creators = ['three_1_x', 'f_integer']
            creators_autoconfig = True

            
        class ItemConstraintsThree(Item):
            model_cls = self.ModelConstraintsThree
        
        ItemConstraintsOne.complete_setup()
        ItemConstraintsThree.complete_setup()
        
        expected_getters = [{'three_1_x'}, {'f_integer'},
                            {'five_1_1'}, {'f_string'}, {'id'}]
        expected_creators = [{'three_1_x'}, {'f_integer'},
                             {'f_text', 'f_string'}]
        
        def sorted_groups(groups):
            subgroups = []
            for subgroup in groups:
                subgroup = list(subgroup)
                subgroup.sort()
                subgroups.append(subgroup)
            subgroups.sort()
        
        self.assertEqual(sorted_groups(ItemConstraintsOne.getters),
                         sorted_groups(expected_getters))
        self.assertEqual(sorted_groups(ItemConstraintsOne.creators),
                         sorted_groups(expected_creators))

        
    def test_model_deleter(self):
        # DeleterXToManyRelationUsed
        for key in ['deleter_selectors', 'deleter_keepers']:
            one_conf = {
                'deleter_selectors': {'f_integer'},
                'deleter_keepers': {'f_string'},
            }
            one_conf[key] = {'two_1_x'}
            with self.assertRaises(DeleterXToManyRelationUsed):
                self.setup_models(
                    one_conf=one_conf,
                    clear_registry=True)
        
        # DeleterSelectorsOrKeepersEmpty
        for key in ['deleter_selectors', 'deleter_keepers']:
            one_conf = {
                'deleter_selectors': {'f_integer'},
                'deleter_keepers': {'f_string'},
            }
            one_conf[key] = []
            with self.assertRaises(DeleterSelectorsOrKeepersEmpty):
                self.setup_models(
                    one_conf=one_conf,
                    clear_registry=True)
                
        # normal setup
        self.setup_models(clear_registry=True)
        self.assertIsNone(self.ItemGeneralOne.metadata['model_deleter'])
        self.assertIsNone(self.ItemGeneralTwo.metadata['model_deleter'])
        
        self.setup_models(
            one_conf = {
                'deleter_selectors': {'f_integer'},
                'deleter_keepers': {'f_string'},
            },
            clear_registry=True)
        self.assertIsNotNone(self.ItemGeneralOne.metadata['model_deleter'])
        model_deleter = self.ItemGeneralOne.metadata['model_deleter']
        self.assertIs(model_deleter.model_cls, self.ItemGeneralOne.model_cls)
        self.assertEqual(model_deleter.selector_fields, {'f_integer'})
        self.assertEqual(model_deleter.keeper_fields, {'f_string'})
        self.assertIsNone(self.ItemGeneralTwo.metadata['model_deleter'])
        
        # default keepers
        self.setup_models(
            one_conf = {
                'deleter_selectors': {'f_integer'},
            },
            clear_registry=True)
        self.assertIsNotNone(self.ItemGeneralOne.metadata['model_deleter'])
        model_deleter = self.ItemGeneralOne.metadata['model_deleter']
        self.assertIs(model_deleter.model_cls, self.ItemGeneralOne.model_cls)
        self.assertEqual(model_deleter.selector_fields, {'f_integer'})
        self.assertEqual(model_deleter.keeper_fields, {'id'})  # primary keys
        self.assertIsNone(self.ItemGeneralTwo.metadata['model_deleter'])
        
        
    def test_model_deleter_unref(self):
        # DeleterXToManyRelationUsed
        for key in ['selectors', 'keepers']:
            with self.assertRaises(DeleterXToManyRelationUsed):
                value = {
                    'selectors': ['f_integer'],
                    'keepers': ['f_integer'],
                }
                value[key] = ['one_1_x']
                self.setup_models(
                    one_conf={
                        'unref_x_to_many': {
                            'two_x_x': value,
                        }
                    },
                    clear_registry=True)
        
        # DeleterSelectorsOrKeepersEmpty
        for key in ['selectors', 'keepers']:
            with self.assertRaises(DeleterSelectorsOrKeepersEmpty):
                value = {
                    'selectors': ['f_integer'],
                    'keepers': ['f_integer'],
                }
                value[key] = []
                self.setup_models(
                    one_conf={
                        'unref_x_to_many': {
                            'two_x_x': value,
                        }
                    },
                    clear_registry=True)
        
        # normal setup
        self.setup_models(clear_registry=True)
        self.assertFalse(self.ItemGeneralOne.metadata['model_unrefs'])
        self.assertFalse(self.ItemGeneralTwo.metadata['model_unrefs'])
        
        self.setup_models(
            one_conf = {
                'unref_x_to_many': {
                    'two_x_x': {
                        'selectors': {'f_string'},
                        'keepers': {'f_text'},
                    },
                    'parent_x_x': {
                        'selectors': {'f_integer'},
                        'keepers': {'f_float'},
                    }
                }
            },
            clear_registry=True)
        self.assertIn('two_x_x',
                      self.ItemGeneralOne.metadata['model_unrefs'])
        self.assertIn('parent_x_x',
                      self.ItemGeneralOne.metadata['model_unrefs'])
        self.assertFalse(self.ItemGeneralTwo.metadata['model_unrefs'])
        
        two_x_x_deleter = \
            self.ItemGeneralOne.metadata['model_unrefs']['two_x_x']
        self.assertIs(two_x_x_deleter.model_cls, self.ItemGeneralTwo.model_cls)
        self.assertEqual(two_x_x_deleter.selector_fields, {'f_string'})
        self.assertEqual(two_x_x_deleter.keeper_fields, {'f_text'})
            
        parent_x_x_deleter = \
            self.ItemGeneralOne.metadata['model_unrefs']['parent_x_x']
        self.assertIs(parent_x_x_deleter.model_cls,
                      self.ItemGeneralOne.model_cls)
        self.assertEqual(parent_x_x_deleter.selector_fields, {'f_integer'})
        self.assertEqual(parent_x_x_deleter.keeper_fields, {'f_float'})
        
        # default keepers
        self.setup_models(
            one_conf = {
                'unref_x_to_many': {
                    'two_x_x': {'f_string'},
                }
            },
            clear_registry=True)
        self.assertIn('two_x_x',
                      self.ItemGeneralOne.metadata['model_unrefs'])
        self.assertFalse(self.ItemGeneralTwo.metadata['model_unrefs'])
        
        two_x_x_deleter = \
            self.ItemGeneralOne.metadata['model_unrefs']['two_x_x']
        self.assertIs(two_x_x_deleter.model_cls, self.ItemGeneralTwo.model_cls)
        self.assertEqual(two_x_x_deleter.selector_fields, {'f_string'})
        self.assertEqual(two_x_x_deleter.keeper_fields, {'id'})  # primary keys
    
    
    def test_remove_null_fields_autoconfig(self):
        self.item_cls_manager.autogenerate = True
        
        # normal defaults ------------------------------------------------------
        self.item_cls_manager.clear()
        
        # normal fields
        ItemConstraintsOne = type('ItemConstraintsOne', (Item,), {
            'model_cls': self.ModelConstraintsOne,
        })
        # relations
        ItemConstraintsTwo = type('ItemConstraintsTwo', (Item,), {
            'model_cls': self.ModelConstraintsTwo,
        })
        # nothing
        ModelConstraintsFive = type('ModelConstraintsFive', (Item,), {
            'model_cls': self.ModelConstraintsFive,
        })
        
        ItemConstraintsOne.complete_setup()
        self.assertEqual(ItemConstraintsOne.remove_null_fields,
                         {'f_text', 'f_string'})
        ItemConstraintsTwo.complete_setup()
        self.assertEqual(ItemConstraintsTwo.remove_null_fields,
                         {'four_x_1'})
        ModelConstraintsFive.complete_setup()
        self.assertEqual(ModelConstraintsFive.remove_null_fields, set())
    
    
        # remove_null_fields_autoconfig off ------------------------------------
        self.item_cls_manager.clear()
        
        # no manual setup
        ItemConstraintsOne = type('ItemConstraintsOne', (Item,), {
            'model_cls': self.ModelConstraintsOne,
            'remove_null_fields_autoconfig': False,
        })
        # manual setup
        ItemConstraintsTwo = type('ItemConstraintsTwo', (Item,), {
            'model_cls': self.ModelConstraintsTwo,
            'remove_null_fields_autoconfig': False,
            'remove_null_fields': {'f_integer', 'f_string'}
        })

        ItemConstraintsOne.complete_setup()
        self.assertEqual(ItemConstraintsOne.remove_null_fields, set())
        ItemConstraintsTwo.complete_setup()
        self.assertEqual(ItemConstraintsTwo.remove_null_fields,
                         {'f_integer', 'f_string'})
        
        # remove_null_fields_autoconfig on -------------------------------------
        self.item_cls_manager.clear()
        
        # no manual setup
        ItemConstraintsOne = type('ItemConstraintsOne', (Item,), {
            'model_cls': self.ModelConstraintsOne,
            'remove_null_fields_autoconfig': True,
        })
        # manual setup
        ItemConstraintsTwo = type('ItemConstraintsTwo', (Item,), {
            'model_cls': self.ModelConstraintsTwo,
            'remove_null_fields_autoconfig': True,
            'remove_null_fields': {'f_integer', 'f_string'}
        })

        ItemConstraintsOne.complete_setup()
        self.assertEqual(ItemConstraintsOne.remove_null_fields,
                         {'f_text', 'f_string'})
        ItemConstraintsTwo.complete_setup()
        self.assertEqual(ItemConstraintsTwo.remove_null_fields,
                         {'f_integer', 'f_string', 'four_x_1'})
        
        # remove_null_fields_autoconfig default --------------------------------
        self.item_cls_manager.clear()
        
        # no manual setup
        ItemConstraintsOne = type('ItemConstraintsOne', (Item,), {
            'model_cls': self.ModelConstraintsOne,
            'remove_null_fields': {}
        })
        # manual setup
        ItemConstraintsTwo = type('ItemConstraintsTwo', (Item,), {
            'model_cls': self.ModelConstraintsTwo,
            'remove_null_fields': {'f_integer', 'f_string'}
        })

        ItemConstraintsOne.complete_setup()
        self.assertEqual(ItemConstraintsOne.remove_null_fields, set())
        ItemConstraintsTwo.complete_setup()
        self.assertEqual(ItemConstraintsTwo.remove_null_fields,
                         {'f_integer', 'f_string'})
        



from save_to_db.exceptions import (RelatedItemNotFound,
                                   MultipleRelatedItemsFound,
                                   ItemAdapterNotFound)
from save_to_db.core.item import Item
from save_to_db.core.item_metaclass import ItemMetaclass
from save_to_db.utils.test_base import TestBase



class TestItemMetaclass(TestBase):
    """ Contains basic tests for an item metaclass.
    
    .. note:
        This test case only tests those features that can be tested without
        using any models (`model_cls` always `None`).
    """
    
    def test_conversions_overwrite(self):
        
        boolean_true_strings = ('boolean', 'true', 'strings',)
        date_formats = ('%m/%d/%Y',)
        
        class ItemA(Item):
            conversions = {
                'boolean_true_strings': boolean_true_strings,
                'date_formats': date_formats,
            }
        
        conversions = ItemA.conversions
        default_conversions = \
            ItemMetaclass.get_default_configuration()['conversions']
        
        # other conversions values must have been inserted
        self.assertEqual(set(conversions.keys()),
                         set(default_conversions.keys()))
        
        overwritten_conversions = {
            'boolean_true_strings': boolean_true_strings,
            'date_formats': date_formats,
        }
        
        for key, default_value in default_conversions.items():
            expected_value = overwritten_conversions.get(key, default_value)
            self.assertEqual(conversions[key], expected_value)
    
    
    def test_conversions_structure_correction(self):
        date_formats = '%d.%m.%y'
        time_formats = '%H%M'
        datetime_formats = '%d.%m.%y %H:%M'
        
        class ItemA(Item):
            conversions = {
                'date_formats': date_formats,
                'time_formats': time_formats,
                'datetime_formats': datetime_formats,
            }
        
        expected = {
            'date_formats': (date_formats,),
            'time_formats': (time_formats,),
            'datetime_formats': (datetime_formats,),
        }
        
        conversions = ItemA.conversions
        default_conversions = \
            ItemMetaclass.get_default_configuration()['conversions']
            
        for key, default_value in default_conversions.items():
            expected_value = expected.get(key, default_value)
            self.assertEqual(conversions[key], expected_value)
        
    
    def test_creators_getters_nullables_structure_correction(self):
        
        class ItemA(Item):
            creators = [['a', 'b'], 'c']
            getters = [['x', 'y'], 'z']
            nullables = ['e', 'f', 'g']
        
        expected_creators = [{'a', 'b'}, {'c'}]
        expected_getters = [{'x', 'y'}, {'z'}]
        expected_nullabels = {'e', 'f', 'g'}
        
        self.assertEqual(ItemA.creators, expected_creators)
        self.assertEqual(ItemA.getters, expected_getters)
        self.assertEqual(ItemA.nullables, expected_nullabels)
    
    
    def test_relations_structure_correction(self):
        
        class ItemA(Item):
            relations = {
                'item_b': 'ItemB',
                'item_c': {
                    'item_cls': 'ItemC',
                },
                'item_d': {
                    'item_cls': 'ItemD',
                    'replace_x_to_many': True,
                },
            }
        
        expected_relations = {
            'item_b': {
                'item_cls': 'ItemB',
                'relation_type': None,
                'replace_x_to_many': False,
                'reverse_key': None,
            },
            'item_c': {
                'item_cls': 'ItemC',
                'relation_type': None,
                'replace_x_to_many': False,
                'reverse_key': None,
            },
            'item_d': {
                'item_cls': 'ItemD',
                'relation_type': None,
                'replace_x_to_many': True,
                'reverse_key': None,
            },
        }
        
        self.assertEqual(ItemA.relations, expected_relations)
    
    
    def test_relations_path_replacements(self):

        #--- proper replacements ---

        class TestRelationsPathReplacements_A(Item):
            relations = {
                'item_b': 'TestRelationsPathReplacements_B',
            }
        
        class TestRelationsPathReplacements_B(Item):
            relations = {
                'item_a': {
                    'item_cls': 'TestRelationsPathReplacements_A',
                }
            }
            
        self.assertFalse(
            TestRelationsPathReplacements_A.metadata['setup_completed'])
        TestRelationsPathReplacements_A.complete_setup()
        relations = TestRelationsPathReplacements_A.relations
        self.assertIs(relations['item_b']['item_cls'],
                      TestRelationsPathReplacements_B)
        self.assertTrue(
            TestRelationsPathReplacements_A.metadata['setup_completed'])
        
        self.assertFalse(
            TestRelationsPathReplacements_B.metadata['setup_completed'])
        TestRelationsPathReplacements_B.complete_setup()
        relations = TestRelationsPathReplacements_B.relations
        self.assertIs(relations['item_a']['item_cls'],
                      TestRelationsPathReplacements_A)
        self.assertTrue(
            TestRelationsPathReplacements_B.metadata['setup_completed'])
        
        #--- related item not found ---
        
        class TestRelationsPathReplacements_C(Item):
            relations = {
                'item': 'TestRelationsPathReplacements_NONEXISTENT',
            }
        
        with self.assertRaises(RelatedItemNotFound):
            TestRelationsPathReplacements_C.complete_setup()
            
        #--- multiple related items found ---
        
        # defining item classes with the same name
        ItemRepeatedOne = type(
            'TestRelationsPathReplacements_REPEATED',
             (Item,), {})
        ItemRepeatedTwo = type('TestRelationsPathReplacements_REPEATED',
             (Item,), {})
        
        class TestRelationsPathReplacements_D(Item):
            relations = {
                    'item': 'TestRelationsPathReplacements_REPEATED',
                }
        
        # getting multiple items from the same module
        ItemRepeatedOne.__module__ = 'a.b.c'
        ItemRepeatedTwo.__module__ = 'a.b.c'
        TestRelationsPathReplacements_D.__module__ = 'a.b.c'
        with self.assertRaises(MultipleRelatedItemsFound):
            TestRelationsPathReplacements_D.complete_setup()
        
        # getting multiple items from different modules
        ItemRepeatedOne.__module__ = 'a.b.c.1'
        ItemRepeatedTwo.__module__ = 'a.b.c.2'
        with self.assertRaises(MultipleRelatedItemsFound):
            TestRelationsPathReplacements_D.complete_setup()
            
        # normal configuration with only one item in the same module and any
        # other items in other modules
        ItemRepeatedOne.__module__ = 'a.b.c'
        TestRelationsPathReplacements_D.complete_setup()
        
    
    def test_item_adapter_not_found(self):
        
        class FakeModelCls(object):
            pass
        
        class ItemA(Item):
            model_cls = FakeModelCls
                
        with self.assertRaises(ItemAdapterNotFound):
            ItemA.complete_setup()
    
    
    def test_unref_x_to_many_structure_correction(self):
        class ItemA(Item):
            unref_x_to_many = {
                'field_1_x_first': {
                    'selectors': ['field_1', 'field_2'],
                    'keepers': ['field_10', 'field_20'],
                },
                'field_1_x_second': {
                    'selectors': ['field_a', 'field_b'],
                },
                'field_1_x_third': ['field_x', 'field_y'],
            }
        ItemA.complete_setup()
        
        expect = {
            'field_1_x_first': {
                'keepers': ['field_10', 'field_20'],
                'selectors': ['field_1', 'field_2']
            },
            'field_1_x_second': {
                'keepers': None,
                'selectors': ['field_a', 'field_b']
            },
            'field_1_x_third': {
                'keepers': None,
                'selectors': ['field_x', 'field_y']
            }
        }
        self.assertEqual(ItemA.unref_x_to_many, expect)
    
    
    def test_remove_null_fields_structure_correction(self):
        class ItemA(Item):
            remove_null_fields = ['field_1', 'field_2']
        ItemA.complete_setup()
        
        self.assertEqual(ItemA.remove_null_fields, {'field_1', 'field_2'})
        
        


from enum import Enum


class RelationType(Enum):
    """ Enumeration class for all possible relation types.
    
    .. note:: See the source code for helper methods that each enumeration has.
    """
    
    # for Python versions prior to 3.6
    def __new__(cls):
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = value
        return obj
    
    ONE_TO_ONE = ()
    ONE_TO_MANY = ()
    MANY_TO_ONE = ()
    MANY_TO_MANY = ()
    
    
    def is_x_to_one(self):
        return self in (RelationType.ONE_TO_ONE, RelationType.MANY_TO_ONE)
    
    def is_x_to_many(self):
        return self in (RelationType.ONE_TO_MANY, RelationType.MANY_TO_MANY)
    
    def is_one_to_x(self):
        return self in (RelationType.ONE_TO_ONE, RelationType.ONE_TO_MANY)
    
    def is_many_to_x(self):
        return self in (RelationType.MANY_TO_ONE, RelationType.MANY_TO_MANY)
    
    def reverse(self):
        if self is RelationType.MANY_TO_ONE:
            return RelationType.ONE_TO_MANY
        
        if self is RelationType.ONE_TO_MANY:
            return RelationType.MANY_TO_ONE
        
        return self


""" This models contains all the signals used by the library. """

from .utils.signal import Signal

all_signals = []
def __create_signal(name):
    global all_signals
    signal = Signal(name)
    all_signals.append(signal)
    return signal


#: Signal is emitted when a persistance process is started.
#: 
#: :param item: The item that was used to persist the data.
#: :param item_structure: Dictionary with single item classes as keys and lists
#:     of single item instances as values.
before_db_persist = __create_signal('before_db_persist')

#: Signal is emitted when a persistance process is completed.
#: 
#: :param items: List of persisted items.
#: :param models: List of lists of persisted models corresponding to the list of
#:     items. One item can have more then one model updated.
after_db_persist = __create_signal('after_db_persist')

#: Signal is emitted during persistence process of an item cannot be used for
#: persistence.
#:
#:    .. note::
#:        If possible, related items will still be persisted.
#:
#: :param item: A dropped item.
#: :param str reason: A short explanation why the item was dropped.
item_dropped = __create_signal('item_dropped')

item_dropped.reason_cannot_create_get_only_mode = \
    'Get only mode with no model matching the item in database'
item_dropped.reason_cannot_create_update_only_mode = \
    'Update only mode with no model matching the item in database'
item_dropped.reason_cannot_create_not_enough_data = \
    'No models matching the item and not enouth data to create new model'



import sqlalchemy as sa
from sqlalchemy.orm import backref, relationship

from ..config import Base



class __CommonColumns(object):
    f_binary = sa.Column(sa.LargeBinary, nullable=True)
    f_boolean = sa.Column(sa.Boolean, nullable=True)
    f_string = sa.Column(sa.String, nullable=True)
    f_text = sa.Column(sa.Text, nullable=True)
    f_integer = sa.Column(sa.Integer, nullable=True)
    f_float = sa.Column(sa.Float, nullable=True)
    f_decimal = sa.Column(sa.Float(asdecimal=True), nullable=True)
    f_date = sa.Column(sa.Date, nullable=True)
    f_time = sa.Column(sa.Time, nullable=True)
    f_datetime = sa.Column(sa.DateTime, nullable=True)


class ModelGeneralOneToOne(Base):
    __tablename__ = 'model_general_one_to_one'
    
    parent_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_general_one.id', ondelete='CASCADE'),
        primary_key=True)
    parent = relationship(
        'ModelGeneralOne',
        foreign_keys=[parent_id])
    
    child_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_general_one.id', ondelete='CASCADE'),
        primary_key=True)
    child = relationship(
        'ModelGeneralOne',
        foreign_keys=[child_id])


class ModelGeneralOne(__CommonColumns, Base):
    __tablename__ = 'model_general_one'
    
    id = sa.Column(sa.Integer, primary_key=True)
    
    #--- self relations ---
    
    # self one-to-one
    parent_1_1_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_general_one.id', ondelete='CASCADE'))
    child_1_1 = relationship(
        'ModelGeneralOne',
        backref=backref('parent_1_1', remote_side=[id]),
        foreign_keys=[parent_1_1_id],
        uselist=False)
    
    # self one-to-many and many-to-one
    parent_x_1_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_general_one.id'))
    child_1_x = relationship(
        'ModelGeneralOne',
        backref=backref('parent_x_1', remote_side=[id]),
        foreign_keys=[parent_x_1_id])
    
    # self many-to-many
    parent_x_x = relationship(
        'ModelGeneralOne',
        secondary='model_general_one_to_one',
        back_populates='child_x_x',
        primaryjoin=(id==ModelGeneralOneToOne.__table__.c.parent_id),
        secondaryjoin=(id==ModelGeneralOneToOne.__table__.c.child_id))
    child_x_x = relationship(
        'ModelGeneralOne',
        secondary='model_general_one_to_one',
        back_populates='parent_x_x',
        primaryjoin=(id==ModelGeneralOneToOne.__table__.c.child_id),
        secondaryjoin=(id==ModelGeneralOneToOne.__table__.c.parent_id))
    
    #--- relations with other ---
    
    # one-to-one
    two_1_1_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_general_two.id', ondelete='CASCADE'))
    two_1_1 = relationship(
        'ModelGeneralTwo',
        back_populates='one_1_1',
        foreign_keys=[two_1_1_id],
        post_update=True)
    
    # one-to-many
    two_1_x = relationship(
        'ModelGeneralTwo',
        foreign_keys='[ModelGeneralTwo.one_x_1_id]',
        back_populates='one_x_1',
        post_update=True)
    
    # many-to-one
    two_x_1_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_general_two.id', ondelete='CASCADE'))
    two_x_1 = relationship(
        'ModelGeneralTwo',
        back_populates='one_1_x',
        foreign_keys=[two_x_1_id],
        post_update=True)
    
    # many-to-many
    two_x_x = relationship(
        'ModelGeneralTwo',
        secondary='model_general_one_to_two',
        back_populates='one_x_x')


class ModelGeneralOneToTwo(Base):
    __tablename__ = 'model_general_one_to_two'
    
    one_id = sa.Column(
        sa.Integer,
        sa.ForeignKey(ModelGeneralOne.id, ondelete='CASCADE'),
        primary_key=True)
    one = relationship(
        ModelGeneralOne,
        foreign_keys=[one_id])
    
    two_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_general_two.id', ondelete='CASCADE'),
        primary_key=True)
    two = relationship(
        'ModelGeneralTwo',
        foreign_keys=[two_id])


class ModelGeneralTwoToTwo(Base):
    __tablename__ = 'model_general_two_to_two'
    
    parent_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_general_two.id', ondelete='CASCADE'),
        primary_key=True)
    parent = relationship(
        'ModelGeneralTwo',
        foreign_keys=[parent_id])
    
    child_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_general_two.id', ondelete='CASCADE'),
        primary_key=True)
    child = relationship(
        'ModelGeneralTwo',
        foreign_keys=[child_id])


class ModelGeneralTwo(__CommonColumns, Base):
    __tablename__ = 'model_general_two'
    
    id = sa.Column(sa.Integer, primary_key=True)

    #--- self relations ---
    
    # self one-to-one
    parent_1_1_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_general_two.id', ondelete='CASCADE'))
    child_1_1 = relationship(
        'ModelGeneralTwo',
        backref=backref('parent_1_1', remote_side=[id]),
        foreign_keys=[parent_1_1_id],
        uselist=False)
    
    # self one-to-many and many-to-one
    parent_x_1_id = sa.Column(
        sa.Integer,
        sa.ForeignKey('model_general_two.id'))
    child_1_x = relationship(
        'ModelGeneralTwo',
        backref=backref('parent_x_1', remote_side=[id]),
        foreign_keys=[parent_x_1_id])
    
    # self many-to-many
    parent_x_x = relationship(
        'ModelGeneralTwo',
        secondary='model_general_two_to_two',
        back_populates='child_x_x',
        primaryjoin=(id==ModelGeneralTwoToTwo.__table__.c.parent_id),
        secondaryjoin=(id==ModelGeneralTwoToTwo.__table__.c.child_id))
    child_x_x = relationship(
        'ModelGeneralTwo',
        secondary='model_general_two_to_two',
        back_populates='parent_x_x',
        primaryjoin=(id==ModelGeneralTwoToTwo.__table__.c.child_id),
        secondaryjoin=(id==ModelGeneralTwoToTwo.__table__.c.parent_id))
    
    
    #--- relations with other ---
    
    # one-to-one
    one_1_1 = relationship(
        ModelGeneralOne,
        back_populates='two_1_1', uselist=False,
        foreign_keys=[ModelGeneralOne.two_1_1_id])
    
    # one-to-many
    one_1_x = relationship(
        ModelGeneralOne,
        foreign_keys=[ModelGeneralOne.two_x_1_id],
        back_populates='two_x_1',
        post_update=True)
    
    # many-to-one
    one_x_1_id = sa.Column(
        sa.Integer,
        sa.ForeignKey(ModelGeneralOne.id, use_alter=True, ondelete='CASCADE'))
    one_x_1 = relationship(
        ModelGeneralOne,
        back_populates='two_1_x',
        foreign_keys=[one_x_1_id],
        post_update=True)
    
    # many-to-many
    one_x_x = relationship(
        ModelGeneralOne,
        secondary='model_general_one_to_two',
        back_populates='two_x_x')
    




from save_to_db.core.item_cls_manager import item_cls_manager
from save_to_db.core.item import Item
from save_to_db.core.merge_policy import MergePolicy
from save_to_db.core.persister import Persister
from save_to_db.core.scope import Scope
from save_to_db.core import signals

from save_to_db.adapters.utils.relation_type import RelationType


import inspect
from .. import ADAPTERS


__cache = {}  # {model_cls: adapter_cls, ...}

def get_adapter_cls(model_or_cls):
    """ Returns database adapter class for given ORM model instance or class.
    
    :param model_or_cls: an ORM model instance or class.
    :returns: An adapter class (subclass of
        :py:class:`~.adapter_base.AdapterBase`) for the model.
    """
    global __cache
    
    cls = model_or_cls
    if not inspect.isclass(cls):
        cls = model_or_cls.__class__

    if cls in __cache:
        return __cache[cls]

    for adapter_cls in ADAPTERS:
        if adapter_cls.is_usable(cls):
            __cache[cls] = adapter_cls
            return adapter_cls

